# Copyright European Space Agency, 2013

from __future__ import division, print_function

from six.moves import zip, map
from six import string_types
import warnings
import os
import sys
import gc
import fnmatch
import time
import json
from datetime import datetime
from collections import OrderedDict

if sys.version_info.major == 2:
    try:
        from numap.NuMap import NuMap
    except ImportError as e:
        print('parallel processing not available (NuMap missing)')
        print(repr(e))
else:
    print('parallel processing not available (NuMap not supported on Python 3 yet)')

import numpy as np

import auromat.fits

from auromat.coordinates.geodesic import wgs84A, wgs84B
from auromat.coordinates.intersection import ellipsoidLineIntersects
from auromat.mapping.mapping import BaseMappingProvider, FileImageMixin,\
    sanitize_data, ArrayImageMixin
from auromat.mapping.astrometry import BaseAstrometryMapping,\
    ImageMaskAstrometryMixin
from auromat.util.decorators import lazy_property, inherit_docs
from auromat.coordinates.ephem import EphemerisCalculator

@inherit_docs
class SpacecraftMappingProvider(BaseMappingProvider):
    def __init__(self, imageSequenceFolder, wcsFolder=None, imageFileExtension=None, timeshift=None, 
                 noradId=None, tleFolder=None, spacetrack=None, altitude=110, maxTimeOffset=3,
                 sequenceInParallel=False):
        """        
        :param imageSequenceFolder: folder path or a list of image file paths
        :param wcsFolder: folder path or a list of wcs file paths;
                          optional if imageSequenceFolder is a folder path and contains
                          the wcs files
        """
        BaseMappingProvider.__init__(self, maxTimeOffset=maxTimeOffset)
        if wcsFolder is None:
            assert not isinstance(imageSequenceFolder, list),\
                   'The wcsFolder parameter is required if imageSequenceFolder is a list'
            wcsFolder = imageSequenceFolder
        if isinstance(imageSequenceFolder, list) and isinstance(wcsFolder, list):
            self.imagePaths = imageSequenceFolder
            self.wcsPaths = wcsFolder
            self._imageFileExtension = os.path.splitext(self.imagePaths[0])[1][1:]
            self._checkEachWcsHasOneImage()
            self._sortByDate()
        elif not isinstance(imageSequenceFolder, list) and not isinstance(wcsFolder, list):
            self.imageSequenceFolder = imageSequenceFolder
            self.wcsFolder = wcsFolder
            self._imageFileExtension = imageFileExtension
            self.reload()            
        else:
            raise ValueError('imageSequenceFolder and wcsFolder must be both path lists or folder paths')
                
        self.timeshift = timeshift
        self.noradId = noradId
        self.tleFolder = tleFolder
        self.spacetrack = spacetrack
        self.altitude = altitude
        
        metadataPath = os.path.join(os.path.dirname(self.imagePaths[0]), 'metadata.json')
        if os.path.exists(metadataPath):
            with open(metadataPath, 'r') as fp:
                self.metadata = json.load(fp, object_hook=_parseDates)
        else:
            self.metadata = None
                
        self._sequenceInParallel = sequenceInParallel
    
    def __len__(self):
        return len(self.wcsPaths)
    
    def reload(self):
        """
        Refresh to current disk state if imageSequenceFolder and wcsFolder
        are folders instead of file path lists.
        """
        wcsFilenames = os.listdir(self.wcsFolder)
        wcsPaths = [os.path.join(self.wcsFolder, f) for f in wcsFilenames]
        self.wcsPaths = fnmatch.filter(wcsPaths, '*.wcs')
        
        imageFilenames = os.listdir(self.imageSequenceFolder)
        imagePaths = [os.path.join(self.imageSequenceFolder, f) for f in imageFilenames]

        try:
            self.imagePaths = fnmatch.filter(imagePaths, '*.' + self.imageFileExtension)
        except ValueError:
            self.imagePaths = []
            self.wcsPaths = []
        
        self._checkEachWcsHasOneImage()
        self._sortByDate()
        
    def _checkEachWcsHasOneImage(self):
        wcsFilenames = map(os.path.basename, self.wcsPaths)
        ids = [os.path.splitext(f)[0] for f in wcsFilenames]
        imageFilenames = list(map(os.path.basename, self.imagePaths))
        imageIds = list(filter(lambda id_: id_ + '.' + self.imageFileExtension in imageFilenames, ids))
        assert len(imageIds) == len(ids), 'image ids: ' + str(imageIds) + '; wcs ids: ' + str(ids)
        self.ids = ids

    def _sortByDate(self):
        dates = {auromat.fits.getShiftedPhotoTime(auromat.fits.readHeader(p)): (p, id_)
                 for p, id_ in zip(self.wcsPaths, self.ids)}
        dates = OrderedDict(sorted(dates.items(), key=lambda k_v: k_v[0]))
        self.dates = dates.keys()
        self.wcsPaths = [p for p,_ in dates.values()]
        self.ids = [id_ for _,id_ in dates.values()]
    
    @property
    def imageFileExtension(self):
        """ e.g. 'jpg' """
        if self._imageFileExtension is None:
            # try to find extension ourselves
            imageFilenames = os.listdir(self.imageSequenceFolder)
            wcsFilenames = fnmatch.filter(os.listdir(self.wcsFolder), '*.wcs')
            if self.imageSequenceFolder == self.wcsFolder:
                imageFilenames = set(imageFilenames) - set(wcsFilenames)
            for wcsFilename in wcsFilenames:
                fileBase = os.path.splitext(wcsFilename)[0]
                matches = fnmatch.filter(imageFilenames, fileBase + '.*')
                if len(matches) == 1:
                    self._imageFileExtension = os.path.splitext(matches[0])[1][1:]
                    break
                elif len(matches) > 1:
                    raise ValueError('Image file extension not given but multiple candidates exist: ' + str(matches))
            if self._imageFileExtension is None:
                raise ValueError('Image file extension could not be determined. Make sure that there exists at least ' +
                                 'one .wcs file and a corresponding image with the same filename base.')
        
        return self._imageFileExtension    
    
    @property
    def range(self):
        return self.dates[0], self.dates[-1]
    
    @property
    def unsolvedIds(self):
        imageFilenames = map(os.path.basename, self.imagePaths)
        imageIds = [os.path.splitext(f)[0] for f in imageFilenames]
        unsolvedIds = filter(lambda id_: id_ not in self.ids, imageIds)
        return sorted(unsolvedIds)
    
    def _getIdxWithOffset(self, date):        
        idx = auromat.utils.findNearest(self.dates, date)
        offset = abs(self.dates[idx]-date).total_seconds()
        return idx, offset
            
    def contains(self, date):
        _, offset = self._getIdxWithOffset(date)
        return offset <= self.maxTimeOffset
    
    def get(self, date):
        idx, offset = self._getIdxWithOffset(date)
        if offset > self.maxTimeOffset:
            raise ValueError('No image found')
        
        identifier = self.ids[idx]
        
        imagePath = os.path.join(self.imageSequenceFolder, identifier + '.' + self.imageFileExtension)
        wcsPath = self.wcsPaths[idx]
        
        if self.metadata:
            metadata = dict(list(self.metadata['sequence_metadata'].items()) + 
                            list(self.metadata['image_metadata'][identifier].items()))
        else:
            metadata = None
        
        mapping = getMapping(imagePath, wcsPath, self.timeshift, 
                             self.noradId, self.tleFolder, self.spacetrack, 
                             altitude=self.altitude,
                             metadata=metadata)
        return mapping
    
    def getById(self, identifier):
        matchedIds = filter(lambda id_: identifier in id_, self.ids)
        assert len(matchedIds) == 1, 'Ambiguous identifier: ' + str(matchedIds)
        identifier = matchedIds[0]
        idx = self.ids.index(identifier)
        return self.get(self.dates[idx])
    
    def getSequence(self, dateBegin=None, dateEnd=None):
        assert dateBegin is None and dateEnd is None, 'Date ranges not supported'
        try:
            self.imageFileExtension
        except ValueError as e:
            warnings.warn(str(e) + ' Returning empty sequence.')
            return []
        
        imagePaths = [os.path.join(self.imageSequenceFolder, id_ + '.' + self.imageFileExtension) for id_ in self.ids]
        
        if self.metadata:
            seqmeta = list(self.metadata['sequence_metadata'].items())
            metadatas = [dict(seqmeta + list(self.metadata['image_metadata'][k].items())) for k in self.ids]
        else:
            metadatas = None
        
        return getMappingSequence(imagePaths, self.wcsPaths, metadatas=metadatas,
                                 timeshift=self.timeshift, noradId=self.noradId, 
                                 tleFolder=self.tleFolder, spacetrack=self.spacetrack, 
                                 altitude=self.altitude,
                                 parallel=self._sequenceInParallel)

@inherit_docs  
class SpacecraftMappingPathProvider(BaseMappingProvider):
    def __init__(self, imagePaths, wcsPaths, metadataPath=None, timeshift=None, 
                 noradId=None, tleFolder=None, spacetrack=None, altitude=110, maxTimeOffset=3,
                 sequenceInParallel=False, fastCenterCalculation=False):
        BaseMappingProvider.__init__(self, maxTimeOffset=maxTimeOffset)
        assert len(imagePaths) == len(wcsPaths)
        self.imagePaths, self.wcsPaths = self._sortByDate(imagePaths, wcsPaths)
                
        self.timeshift = timeshift
        self.noradId = noradId
        self.tleFolder = tleFolder
        self.spacetrack = spacetrack
        self.altitude = altitude
        self.sequenceInParallel = sequenceInParallel
        self.fastCenterCalculation = fastCenterCalculation
        
        if metadataPath and os.path.exists(metadataPath):
            with open(metadataPath, 'r') as fp:
                self.metadata = json.load(fp, object_hook=_parseDates)
        else:
            self.metadata = None
             
    def __len__(self):
        return len(self.wcsPaths)

    @staticmethod
    def _sortByDate(imagePaths, wcsPaths):
        def date(wcsPath_imagePath):
            wcsPath = wcsPath_imagePath[0]
            wcsHeader = auromat.fits.readHeader(wcsPath)
            return auromat.fits.getPhotoTime(wcsHeader)
        paths = sorted(zip(wcsPaths, imagePaths), key=date)
        wcsPaths = [wcsPath for wcsPath, _ in paths]
        imagePaths = [imagePath for _, imagePath in paths]
        return imagePaths, wcsPaths
    
    @property
    def imageFileExtension(self):
        return os.path.splitext(self.imagePaths[0])[1][1:]
    
    @property
    def range(self):
        fromDate = getMapping(self.imagePaths[0], self.wcsPaths[0]).photoTime
        toDate = getMapping(self.imagePaths[-1], self.wcsPaths[-1]).photoTime
        return fromDate, toDate
    
    def contains(self, date):
        raise NotImplementedError
        
    def get(self, date):
        # TODO implement provider access by date
        raise NotImplementedError
    
    def getById(self, identifier):
        raise NotImplementedError
    
    def getSequence(self, dateBegin=None, dateEnd=None):
        assert dateBegin is None and dateEnd is None, 'Date ranges not supported'
        
        if self.metadata:
            keys = [os.path.splitext(os.path.basename(p))[0] for p in self.imagePaths]
            seqmeta = list(self.metadata['sequence_metadata'].items())
            metadatas = [dict(seqmeta + list(self.metadata['image_metadata'][k].items())) for k in keys]
        else:
            metadatas = None
        
        return getMappingSequence(self.imagePaths, self.wcsPaths,
                                 timeshift=self.timeshift, noradId=self.noradId, 
                                 tleFolder=self.tleFolder, spacetrack=self.spacetrack, 
                                 altitude=self.altitude,
                                 parallel=self.sequenceInParallel, 
                                 fastCenterCalculation=self.fastCenterCalculation,
                                 metadatas=metadatas)

def _getMappingSequenceArgs(imagePathsOrArrays, wcsPaths, timeshift=None, 
                            noradId=None, tleFolder=None, spacetrack=None, altitude = 110,
                            fastCenterCalculation=False, metadatas=None):   
    if not metadatas:
        metadatas = [{}] * len(wcsPaths)
        
    return (dict(imagePathOrArray=imagePathOrArray, wcsPathOrHeader=wcsPath,
                 timeshift=timeshift, noradId=noradId,
                 tleFolder=tleFolder, spacetrack=spacetrack, 
                 altitude=altitude,
                 fastCenterCalculation=fastCenterCalculation,
                 metadata=metadata)
            for imagePathOrArray, wcsPath, metadata in zip(imagePathsOrArrays, wcsPaths, metadatas))
        
def getMappingSequence(imagePathsOrArrays, wcsPaths, metadatas=None, timeshift=None, 
                       noradId=None, tleFolder=None, spacetrack=None, altitude = 110,
                       parallel=False, fastCenterCalculation=False):
    """
    Returns a generator of SpacecraftMapping objects
    for all images in 'imageSequenceFolder' which have a solution in 'wcsFolder'.
    The order corresponds to the sorted filenames in 'wcsFolder'.
    
    :param iterable imagePathsOrArrays:
    :param list wcsPaths:
    """
    mappingArgsArr = _getMappingSequenceArgs(imagePathsOrArrays, wcsPaths, 
                                             timeshift, noradId, tleFolder, spacetrack, 
                                             altitude, fastCenterCalculation,
                                             metadatas=metadatas)
    
    if parallel:
        return _getMappingsParallel(mappingArgsArr)
    else:
        def mappingFromKw(kw):
            mapping = getMapping(**kw)
            # see _getMappingsParallel
            gc.collect()
            return mapping
        return map(mappingFromKw, mappingArgsArr)

def _getMappingsParallel(mappingArgsArr):
    # Each worker process takes 2-4GiB!
    # We use only one worker as the main process is usually slower in consuming
    # the mappings.
    workerCount = 1
    
    # TODO use iterator class instead of yield to conserve memory 
    # (local variable 'mapping' holds on to reference and is only released on next iteration)
    
    mappings = NuMap(_getCalculatedMappingFromArgs, mappingArgsArr,
                     worker_type='process', worker_num=workerCount, buffer=workerCount)
    try:
        mappings.start()
        for mapping in mappings:
            yield mapping
            
            # Due to some reference cycles there are numpy arrays which don't
            # get freed implicitly. As the arrays we work with are quite huge
            # this adds up quickly. The problem is that currently the garbage collector
            # doesn't know about the real size of the numpy arrays (as they are C extension
            # types and there is no API yet for communicating the real size to the Python
            # interpreter). Therefore the thresholds for triggering a garbage collection
            # are seldomly reached and instead we consume more and more memory and eventually
            # run out of it. To fight against this, we manually run a collection to force
            # freeing up native memory.
            gc.collect()
    finally:
        mappings.stop(ends=[0])
   
def _getCalculatedMappingFromArgs(kwargs):
    """
    A helper function which gets a mapping and forces the
    calculation of its (lazy) properties.
    See getMappingSequence().
    """
    try:
        os.nice(10) # only on UNIX systems
    except:
        pass
    mapping = getMapping(**kwargs)
    # force calculation within worker process
    mapping.boundingBox
    mapping.elevation
    return mapping
            

def getMapping(imagePathOrArray, wcsPathOrHeader, timeshift=None, noradId=None, tleFolder=None, spacetrack=None, 
               altitude=110, fastCenterCalculation=False, metadata=None, nosanitize=False,
               identifier=None):
    """
    If timeshift is None, then the wcs header is first checked for a shifted timestamp
    and corresponding spacecraft position. In case no shifted timestamp exists, the
    wcs header is checked for the original timestamp and spacecraft position. If only the
    timestamp exists (which may be the case for externally produced wcs files), the 
    spacecraft position is calculated from two-line elements. If the latter applies or
    'timeshift' is given, then tleFolder must be given.
    If the tleFolder doesn't contain a %noradid%.tle file, then spacetrack is used to download
    the data (or an error is raised if spacetrack is None). The NORAD ID is determined from
    the noradId parameter, or if that is None from the wcs header. If in the latter case 
    the wcs header doesn't contain the NORAD ID, then the ISS ID (25544) is used as a default and a
    warning is printed. 

    :param imagePathOrArray:
    :param wcsPathOrHeader:
    :param datetime.timedelta timeshift: if set, overrides the shifted timestamp stored in the wcs headers
    :param noradId: if set, overrides the NORAD ID stored in the wcs headers
    :param tleFolder: folder containing TLE files named noradid.tle
    :param spacetrack: a Spacetrack class instance
    :param altitude:
    :rtype: BaseSpacecraftMapping
    """
    wcsHeader, photoTime, originalPhotoTime, cameraPosGCRS = \
        _prepareMappingParams(wcsPathOrHeader, timeshift, noradId, tleFolder, spacetrack)
    
    isImageArray = not isinstance(imagePathOrArray, string_types)
    isWcsHeader = not isinstance(wcsPathOrHeader, string_types)
        
    if identifier is None:
        if not isImageArray:
            identifier = os.path.splitext(os.path.basename(imagePathOrArray))[0]
        elif not isWcsHeader:
            identifier = os.path.splitext(os.path.basename(wcsPathOrHeader))[0]
    
    if isImageArray:
        cls = ArraySpacecraftMapping        
    else:
        cls = FileSpacecraftMappingUnsanitized if nosanitize else FileSpacecraftMapping
        
    mapping = cls(wcsHeader, altitude, imagePathOrArray, cameraPosGCRS, photoTime, 
                  identifier, metadata,
                  originalPhotoTime=originalPhotoTime, 
                  fastCenterCalculation=fastCenterCalculation)            
    return mapping 

def _prepareMappingParams(wcsPathOrHeader, timeshift=None, noradId=None, tleFolder=None, spacetrack=None):
    if noradId is not None:
        noradId = int(noradId)
    
    if isinstance(wcsPathOrHeader, string_types):
        fitsWcsHeader = auromat.fits.readHeader(wcsPathOrHeader)
    else:
        fitsWcsHeader = wcsPathOrHeader
    
    originalPhotoTime = auromat.fits.getPhotoTime(fitsWcsHeader)
    if originalPhotoTime is None:
        raise ValueError('DATE-OBS missing in FITS header')

    if timeshift is not None:
        photoTime = originalPhotoTime + timeshift
        cameraPosGCRS = None
    else:
        cameraPosGCRS, photoTime_, _ = auromat.fits.getShiftedSpacecraftPosition(fitsWcsHeader)
        if cameraPosGCRS is not None:
            photoTime = photoTime_
        else:        
            photoTime = originalPhotoTime
            cameraPosGCRS, _ = auromat.fits.getSpacecraftPosition(fitsWcsHeader)
            if cameraPosGCRS is None:
                warnings.warn('Spacecraft position is missing in FITS header, will recalculate')
    
    if cameraPosGCRS is None:
        if noradId is None:
            noradId = auromat.fits.getNoradId(fitsWcsHeader)
            if noradId is None:
                warnings.warn('NORAD ID is missing in FITS header, assuming ISS (25544)')
                noradId = 25544
        
        if tleFolder is None:
            raise ValueError('You need to specify tleFolder to calculate spacecraft positions')

        tleFilePath = os.path.join(tleFolder, str(noradId) + '.tle')
        if os.path.exists(tleFilePath):
            # the EphemerisCalculator doesn't need to be cached, fast enough (0.007s)
            ephemCalculator = EphemerisCalculator(tleFilePath)
            if not ephemCalculator.contains(photoTime):
                if spacetrack is None:
                    raise ValueError('Please update ' + tleFilePath + ' or ' + 
                                     'supply a spacetrack instance for automatic download')
                spacetrack.updateTLEsFor(noradId, tleFilePath, photoTime)
                ephemCalculator = EphemerisCalculator(tleFilePath)
            
        elif spacetrack is not None:
            spacetrack.updateTLEsFor(noradId, tleFilePath, photoTime)
            ephemCalculator = EphemerisCalculator(tleFilePath)
            
        else:
            raise ValueError('Please put ' + str(noradId) + '.tle inside ' + tleFolder + ' or ' + 
                             'supply a spacetrack instance for automatic download')
            
        cameraPosGCRS = ephemCalculator(photoTime)
    
    return fitsWcsHeader, photoTime, originalPhotoTime, cameraPosGCRS 

class BaseSpacecraftMapping(BaseAstrometryMapping):
    """
    A mapping which is based on having a camera in/on a spacecraft looking both on earth 
    and the stars and where no exact camera pointing is known.
    The stars were then used to derive a WCS definition with which it is possible to
    calculate the direction vector of each pixel.
    """

    def __init__(self, wcsHeader, alti, cameraPosGCRS, photoTime, identifier, metadata=None, 
                 originalPhotoTime=None, fastCenterCalculation=False):
        BaseAstrometryMapping.__init__(self, wcsHeader, alti, cameraPosGCRS, photoTime, 
                                       identifier, metadata, 
                                       fastCenterCalculation=fastCenterCalculation)
        if originalPhotoTime is None:
            originalPhotoTime = photoTime
        self._originalPhotoTime = originalPhotoTime
            
    @property
    def originalPhotoTime(self):
        return self._originalPhotoTime
        
    @lazy_property
    def intersectsEarth(self):
        """
        Returns a boolean array indicating whether a pixel center intersects with the earth.
        """
        direction = self.cameraToPixelCenterDirection
        t0 = time.time()
        intersectsEarth = ellipsoidLineIntersects(wgs84A, wgs84B,
                                                  self.cameraPosGCRS,
                                                  direction.reshape(-1,3))
        print('intersectsEarth:', time.time()-t0, 's')
        intersectsEarth = intersectsEarth.reshape(self.cameraToPixelCenterDirection.shape[0],
                                                  self.cameraToPixelCenterDirection.shape[1])
        return intersectsEarth
    
    def isConsistent(self, starPxCoords=None):
        """
        Checks if the photo timestamp and astrometric solution used for mapping is plausible by analysing
        the mapping result.
        
        Note that in general there are virtually no false solves when using astrometry.net.
        
        :param starPxCoords: array of shape (n,2) containing x,y pixel coordinates of stars which have been
                             used for obtaining an astrometry solution;
                             for astrometry.net, the quad stars can be used for this purpose,
                             see auromat.solving.readQuadMatch()
        :rtype: True if consistent, False if not
        """
        if np.all(self.intersectsEarth):
            # Although we solved the image using stars, every pixel intersects
            # with the modelled earth. Thus, the camera position is such
            # that the camera would look directly at the earth, with no starfield in the image.
            # The timestamp and/or astrometric solution must be wrong. 
            return False
        elif not np.any(self.intersectsEarth):
            # No pixel intersects with the modelled earth. As we assume that the images always contain
            # a part of the earth, this must again be a wrong timestamp and/or astrometric solution.
            return False
        
        if starPxCoords is not None:
            starCoveredByEarth = self.intersectsEarth[starPxCoords[:,1],starPxCoords[:,0]]
            if np.any(starCoveredByEarth):
                # There is at least one star used for the astrometry solution which would
                # be covered by the modelled earth. Therefore, the timestamp and/or astrometric
                # solution must be wrong.
                return False
        
        return True

@inherit_docs
class FileSpacecraftMappingUnsanitized(ImageMaskAstrometryMixin, FileImageMixin, BaseSpacecraftMapping):
    """
    .. warning:: Consider using FileSpacecraftMapping instead of this class.
                Masking is not supported here.
    
    The purpose of this class is to access certain properties in a very efficient way by
    skipping any sanitization.
    See auromat.test.draw_test.testParallelsMeridiansPlotOptimized for an example usage of this
    behaviour.
    """
    
    def __init__(self, wcsHeader, alti, imagePath, cameraPosGCRS, photoTime, identifier, metadata=None,
                 originalPhotoTime=None, fastCenterCalculation=False):
        ImageMaskAstrometryMixin.__init__(self)
        FileImageMixin.__init__(self, imagePath)
        BaseSpacecraftMapping.__init__(self, wcsHeader, alti, cameraPosGCRS, photoTime, identifier, metadata, 
                                       originalPhotoTime=originalPhotoTime,
                                       fastCenterCalculation=fastCenterCalculation)
    
    def createMasked(self, centerMask):
        raise RuntimeError('Masking is not supported for unsanitized mappings, ' +
                           'please use nosanitize=False in getMapping()')

FileSpacecraftMapping = sanitize_data(FileSpacecraftMappingUnsanitized)

@sanitize_data
@inherit_docs
class ArraySpacecraftMapping(ImageMaskAstrometryMixin, ArrayImageMixin, BaseSpacecraftMapping):   
    """
    Like FileSpacecraftMapping but accepts an RGB image array instead of an image file path.
    """
    def __init__(self, wcsHeader, alti, img, cameraPosGCRS, photoTime, identifier, metadata=None,
                 originalPhotoTime=None, fastCenterCalculation=False):
        ImageMaskAstrometryMixin.__init__(self)
        ArrayImageMixin.__init__(self, img)
        BaseSpacecraftMapping.__init__(self, wcsHeader, alti, cameraPosGCRS, photoTime, identifier, metadata, 
                                       originalPhotoTime=originalPhotoTime,
                                       fastCenterCalculation=fastCenterCalculation)
    
isoDateFormat = '%Y-%m-%dT%H:%M:%S.%f'

def _parseDates(dic):
    keys = {'date'} & set(dic.keys())
    for k in keys:
        dic[k] = datetime.strptime(dic[k], isoDateFormat)
    return dic


# Copyright European Space Agency, 2013

from __future__ import absolute_import, print_function

from six import reraise
from six.moves.urllib.error import HTTPError
from six.moves.urllib.request import urlopen
import shutil
import os
import json
import sys

# monkey-patch HTTPError and add a __repr__ method
# so that it doesn't display as 'HTTPError()' but like __str__
# as e.g. 'HTTP Error 500: msg'
HTTPError.__repr__ = HTTPError.__str__

def urlResponseCode(url):
    """
    Return the response code of the server without downloading the
    actual data.
    """
    try:
        code = _urlResponseCode(url)
    except: # try again once in case of network problems
        code = _urlResponseCode(url)
    return code

def _urlResponseCode(url):
    try:
        connection = urlopen(url)
        code = connection.getcode()
        connection.close()
    except HTTPError as e:
        code = e.getcode()
    return code

class DownloadError(Exception):
    pass

def downloadFile(url, path, unifyErrors=True):
    """
    Download a single resource and store it as file to disk.
    On download errors (except 404), the download is retried once,
    after that an exception is raised.
    """
    def saveToDisk(req):
        tmpPath = path + '.tmp'
        with open(tmpPath,'wb') as fp:
            shutil.copyfileobj(req, fp)
        os.rename(tmpPath, path)
    downloadResource(url, saveToDisk, unifyErrors=unifyErrors)
    
def downloadJSON(url, unifyErrors=True, data=None, **kw):
    """
    Parse and return the JSON document at the given URL.
    Any additional keywords are given to json.load unchanged.    
    """
    def asjson(req):
        return json.load(req, **kw)
    return downloadResource(url, asjson, data=data, unifyErrors=unifyErrors)

def downloadResource(url, fn, data=None, unifyErrors=True):
    """
    Download a single resource and call `fn` on it.
    On download errors (except 404), the download is retried once,
    after that an exception is raised.
    """
    retry = False
    try:
        return _downloadResource(url, fn, data=data)
    except HTTPError as e:
        if e.code == 404:
            if unifyErrors:
                raise DownloadError(e)
            else:
                raise
        else:
            retry = True
    except IOError as e:
        if unifyErrors:
            raise DownloadError(e)
        else:
            raise
    except:
        retry = True
    
    if retry:
        print('download error, retrying once')
        try:
            return _downloadResource(url, fn, data=data)
        except:
            if unifyErrors:
                _, e, tb = sys.exc_info()
                new_exc = DownloadError('{}: {}'.format(e.__class__.__name__, e))
                reraise(new_exc.__class__, new_exc, tb)
            else:
                raise
    
def _downloadResource(url, fn, data=None):
    print('downloading', url, end='')
    try:        
        req = urlopen(url, data=data) # throws also on 404
        res = fn(req)
        print('-> done')
        return res
    except Exception as e:
        # 404, network problem, IO error, ...
        print('->', str(e))
        raise
        
def downloadFiles(urls, paths, retFailures=False):
    """
    Downloads multiple resources and stores them to disk at the given `paths`,
    ignoring already existing files
    on disk. On download errors (except 404), the download is retried once.
    If retFailures is False, then True is returned if all files
    could be downloaded successfully, otherwise False.
    If retFailures is True, then a tuple (bool, failures) is
    returned which additionally contains the urls (with exceptions)
    that couldn't be downloaded.
    """
    failures = []
    for url, path in zip(urls, paths):
        if os.path.exists(path):
            continue
        try:
            downloadFile(url, path, unifyErrors=False)
        except Exception as e:
            failures.append((url, e))
                
    if retFailures:
        return len(failures) == 0, failures
    else:
        return len(failures) == 0



# Copyright European Space Agency, 2013

from __future__ import division

import os.path
import unittest
from nose.plugins.attrib import attr
import numpy as np
import numpy.ma as ma
from numpy.testing.utils import assert_array_almost_equal,\
    assert_array_equal, assert_equal
    
from auromat.coordinates.geodesic import wgs84A, wgs84B
from auromat.utils import unitVectors
from auromat.coordinates.transform import geodetic2Ecef
from auromat.coordinates.intersection import sphereLineIntersection,\
    _ellipsoidLineIntersection_np, _ellipsoidLineIntersection_ne,\
    _ellipsoidLineIntersects_np, _ellipsoidLineIntersects_ne
from auromat.mapping import miracle, spacecraft

ellipsoidLineIntersectionFns = [_ellipsoidLineIntersection_np, _ellipsoidLineIntersection_ne]
ellipsoidLineIntersectsFns = [_ellipsoidLineIntersects_np, _ellipsoidLineIntersects_ne]

class Test(unittest.TestCase):

    def testSphereLineIntersection(self):
        sphereRadius = 2
        lineOrigin = [0,3,0]
        lineDirection = [0,-1,0]
        point = sphereLineIntersection(sphereRadius, lineOrigin, lineDirection)
        
        assert_equal(point, [0,2,0])
        
    def testSphereLineIntersectionArray(self):
        sphereRadius = 2
        lineOrigin = [0,3,0]
        lineDirection = unitVectors([[0,-1,0],[-1,-1,0]])
        points = sphereLineIntersection(sphereRadius, lineOrigin, lineDirection)
        
        assert_equal(points, [[0,2,0],[np.nan,np.nan,np.nan]])
        
    def testEllipsoidLineIntersection(self):
        for ellipsoidLineIntersection in ellipsoidLineIntersectionFns:
            p1 = np.array(geodetic2Ecef(np.deg2rad(30),np.deg2rad(60),0))
            p2 = np.array(geodetic2Ecef(np.deg2rad(-30),np.deg2rad(-60),0))
            
            i1 = ellipsoidLineIntersection(wgs84A, wgs84B, p1, [p1-p2], directed=False)
    
            assert_array_almost_equal(i1, [p1])
        
    def testEllipsoidLineIntersectionSphere(self):
        for ellipsoidLineIntersection, ellipsoidLineIntersects in zip(ellipsoidLineIntersectionFns, ellipsoidLineIntersectsFns):
            a = b = 2
            lineOrigin = [0,3,0]
            lineDirection = [[0,-1,0],[0,-1,0],[-1,-1,0]]
            
            points = ellipsoidLineIntersection(a, b, lineOrigin, lineDirection)
            assert_equal(points, [[0,2,0],[0,2,0],[np.nan,np.nan,np.nan]])
            
            intersects = ellipsoidLineIntersects(a, b, lineOrigin, lineDirection)
            assert_equal(intersects, [True, True, False])
        
    def testDirectedIntersection(self):
        for ellipsoidLineIntersection, ellipsoidLineIntersects in zip(ellipsoidLineIntersectionFns, ellipsoidLineIntersectsFns):
            r = 1
            origin = [2,0,0]
            direction = [[1,0,0]]
            intersection = [[1,0,0]]
            noIntersection = [[np.nan, np.nan, np.nan]]
            
            res = sphereLineIntersection(r, origin, direction, directed=False)
            assert_array_equal(res, intersection)
            
            res = sphereLineIntersection(r, origin, direction, directed=True)
            assert_array_equal(res, noIntersection)
                    
            res = ellipsoidLineIntersection(r, r, origin, direction, directed=False)
            assert_array_equal(res, intersection)
            
            intersects = ellipsoidLineIntersects(r, r, origin, direction, directed=False)
            assert_equal(intersects, [True])        
            
            res = ellipsoidLineIntersection(r, r, origin, direction, directed=True)
            assert_array_equal(res, noIntersection)
            
            intersects = ellipsoidLineIntersects(r, r, origin, direction, directed=True)
            assert_equal(intersects, [False])
            
            origin2 = [-2,0,0]
            intersection2 = [[-1,0,0]]
            res = sphereLineIntersection(r, origin2, direction, directed=False)
            assert_array_equal(res, intersection2)
            
            res = sphereLineIntersection(r, origin2, direction, directed=True)
            assert_array_equal(res, intersection2)
            
            direction2 = [[-1,0,0]]
            res = sphereLineIntersection(r, origin2, direction2, directed=False)
            assert_array_equal(res, intersection2)
            
            res = sphereLineIntersection(r, origin2, direction2, directed=True)
            assert_array_equal(res, noIntersection)
            
    def testDirectedIntersectionFromInside(self):
        # line origin is inside sphere/ellipsoid
        
        for ellipsoidLineIntersection, ellipsoidLineIntersects in zip(ellipsoidLineIntersectionFns, ellipsoidLineIntersectsFns):
            r = 2
            origin = [1,0,0]
            direction = [[1,0,0]]
            intersection = [[2,0,0]]
            
            res = sphereLineIntersection(r, origin, direction, directed=False)
            assert_array_equal(res, intersection)
            
            res = sphereLineIntersection(r, origin, direction, directed=True)
            assert_array_equal(res, intersection)
                    
            res = ellipsoidLineIntersection(r, r, origin, direction, directed=False)
            assert_array_equal(res, intersection)
            
            intersects = ellipsoidLineIntersects(r, r, origin, direction, directed=False)
            assert_equal(intersects, [True])        
            
            res = ellipsoidLineIntersection(r, r, origin, direction, directed=True)
            assert_array_equal(res, intersection)
            
            intersects = ellipsoidLineIntersects(r, r, origin, direction, directed=True)
            assert_equal(intersects, [True])
            
            intersection2 = [[-2,0,0]]           
            direction2 = [[-1,0,0]]
            res = sphereLineIntersection(r, origin, direction2, directed=False)
            assert_array_equal(res, intersection)
            
            res = sphereLineIntersection(r, origin, direction2, directed=True)
            assert_array_equal(res, intersection2)
        
    @attr('slow')       
    def testIntersectionBug(self):
        """
        An optimized code path led to choosing the intersection points on the
        other side of the earth for a particular mapping. This mapping
        has (wrongly appearing) intersection points at the top of the image, 
        and these errorneous points led the optimized code path to choose
        the wrong side of the earth, at least from the point of view of
        the correct intersection points at the lower half of the image.
        Thus, this issue will disappear once the actual bug is solved.
        """
        m = _getISSMapping()
        dists = m.distance
        assert dists[-1,2000] < dists[-100,2000]
    
    @attr('slow') 
    def testIntersectionBug2(self):
        """
        For the sequence beginning with ISS029-E-8492 there are intersection
        areas at the top of the image which should not be there.
        These intersections occur because the other ends of the rays actually intersect
        with the earth "behind" the camera. This happens because we intersect an
        infinite line with the earth, instead of a line beginning at the camera position
        towards the earth.
        """
        m = _getISSMapping()
        # check that there is intersection at the bottom
        assert m.lats[-1,2000] is not ma.masked
        
        # check that there is NO intersection at the top
        assert m.lats[0,2000] is ma.masked

def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)

def _getISSMapping():
    imagePath = getResourcePath('ISS029-E-8492.jpg')
    wcsPath = getResourcePath('ISS029-E-8492.wcs')
    mapping = spacecraft.getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    return mapping

def _getMiracleMapping():
    imagePath = getResourcePath('SOD120304_171900_557_1000.jpg')
    mapping = miracle.getMapping(imagePath)
    return mapping



# Copyright European Space Agency, 2013

"""
This module georeferences images taken from a spacecraft (e.g.
a camera onboard the ISS) and looking down at earth 
using the starfield in the images.

The resulting calibrations can be used with the corresponding
:mod:`auromat.mapping.spacecraft` module to compute geodetic coordinates
for the image pixels.
"""

from __future__ import division, absolute_import, print_function

from six.moves import zip
import os

import auromat.fits
from auromat.coordinates.spacetrack import Spacetrack
from auromat.coordinates.ephem import EphemerisCalculator
from auromat.util.image import readExifTime
from auromat.solving.masking import maskStarfield
import auromat.solving.solving
from auromat.util.os import makedirs

supportedImageFormats = ('.jpg', '.jpeg', '.png')

def solve(imagePath, wcsPath, tleFolder, spacetrackUser, spacetrackPassword, noradId, 
          debugOutputFolder=None, overwrite=False, maskingFn=maskStarfield, solveTimeout=60*5, 
          noAstrometryPlots=False, channel=None, sigma=None, 
          oddsToSolve=None, verbose=False):
    """
    Return True if the image could be georeferenced, otherwise False.
    
    :param wcsPath: path of output file containing calibration WCS headers
    :param tleFolder: folder where TLE files are/shall be stored
    :param spacetrackUser, spacetrackPassword: 
        Credentials to space-track.org used to download TLE data which in turn are
        used to calculte the spacecraft position at the time the image was taken.
        These can be None if the `tleFolder` contains relevant data already.
    :param noradId: NORAD ID of the spacecraft
    :param debugOutputFolder: folder to store debug images in from masking and astrometry
    :param overwrite: True overwrites an existing wcs file, False raises an exception
    :param maskingFn: the function to use for masking the starfield in the image
    :param solveTimeout: maximum time in seconds to spend for one astrometry run
                         (note that multiple parameter combinations are tried and each
                          counts as a run)
    :param noAstrometryPlots: whether to let astrometry.net produce star overlay plots
                              (only relevant when `debugOutputFolder` is given)
    :param channel: R,G,B, or None (average of all channel), the channel to use for masking
                    and astrometry
    :param oddsToSolve: default 1e9, see astrometry.net docs
    :return: True if astrometry was successful (in which case a file at `wcsPath` was written),
             otherwise False.
    :rtype: bool
    """
    wcsHeaders = _solveImages([imagePath], tleFolder, spacetrackUser, spacetrackPassword, noradId, 
                              maskingFn, debugOutputFolder, solveTimeout=solveTimeout,
                              noAstrometryPlots=noAstrometryPlots,
                              channel=channel, sigma=sigma, oddsToSolve=oddsToSolve, verbose=verbose)
    wcsHeader = next(wcsHeaders)
    if wcsHeader is None:
        return False
    auromat.fits.writeHeader(wcsPath, wcsHeader, overwrite=overwrite)
    return True

def solveSequence(imageSequenceFolder, wcsFolder, tleFolder, spacetrackUser, spacetrackPassword, noradId, 
                  debugOutputFolder=None, parallel=True, maxWorkers=None, 
                  solveTimeout=60*5, maskingFn=maskStarfield, oddsToSolve=None,
                  noAstrometryPlots=False, channel=None, sigma=None, verbose=False):
    """
    Returns a generator containing (imagePath, wcsPath) tuples of successfully solved images.
    This allows to execute actions directly after each solve.
    Images which are already solved are ignored.
    
    See :func:`solve` for parameters.
    """    
    fileNames = os.listdir(imageSequenceFolder)
    imageFileNames = sorted(filter(lambda f: f.endswith(supportedImageFormats), fileNames))
    imagePaths = map(lambda f: os.path.join(imageSequenceFolder, f), imageFileNames)
    return solveImages(imagePaths, wcsFolder, tleFolder, spacetrackUser, spacetrackPassword, noradId, 
                       debugOutputFolder, parallel=parallel, maxWorkers=maxWorkers,
                       solveTimeout=solveTimeout,
                       maskingFn=maskingFn, noAstrometryPlots=noAstrometryPlots,
                       oddsToSolve=oddsToSolve,
                       channel=channel, sigma=sigma, verbose=verbose)
    
def solveImages(imagePaths, wcsFolder, tleFolder, spacetrackUser, spacetrackPassword, noradId, 
                debugOutputFolder=None, parallel=True, maxWorkers=None, 
                solveTimeout=60*5, maskingFn=maskStarfield,
                returnUnsolved=False, noAstrometryPlots=False, oddsToSolve=None,
                channel=None, sigma=None, verbose=False):
    """
    Returns a generator containing (imagePath, wcsPath) tuples of successfully solved images.
    This allows to execute actions directly after each solve.
    Images which are already solved are ignored.
    
    See :func:`solve` for parameters.
    """
    makedirs(wcsFolder)
    unsolvedImagePaths = []
    unsolvedWcsPaths = []
    for imagePath in imagePaths:
        fileName = os.path.basename(imagePath)
        fileBase = os.path.splitext(fileName)[0]
        wcsPath = os.path.join(wcsFolder, fileBase + '.wcs')
        if not os.path.exists(wcsPath):
            unsolvedImagePaths.append(imagePath)
            unsolvedWcsPaths.append(wcsPath)
        
    wcsHeaders = _solveImages(unsolvedImagePaths, 
                              tleFolder, spacetrackUser, spacetrackPassword, 
                              noradId, maskingFn,
                              debugOutputFolder=debugOutputFolder, 
                              parallel=parallel, maxWorkers=maxWorkers, 
                              solveTimeout=solveTimeout,
                              noAstrometryPlots=noAstrometryPlots, oddsToSolve=oddsToSolve,
                              channel=channel, sigma=sigma, verbose=verbose)
    
    for (imagePath, wcsPath, wcsHeader) in zip(unsolvedImagePaths, unsolvedWcsPaths, wcsHeaders):
        if wcsHeader is not None:
            auromat.fits.writeHeader(wcsPath, wcsHeader)
            yield (imagePath, wcsPath)
        elif returnUnsolved:
            yield (imagePath, None)

def _solveImages(imagePaths, tleFolder, spacetrackUser, spacetrackPassword, noradId, maskingFn, 
                 debugOutputFolder=None, parallel=True, solveTimeout=60*5, maxWorkers=None, 
                 noAstrometryPlots=False, channel=None, sigma=None, oddsToSolve=None, 
                 verbose=False):
    """
    Returns a generator containing the wcsHeader (or None) for each image.
    """    
    noradId = int(noradId)
            
    originalPhotoTimes = list(map(readExifTime, imagePaths))
    latestPhotoTime = max(originalPhotoTimes)
    
    tleFilePath = os.path.join(tleFolder, str(noradId) + '.tle')
    spacetrack = Spacetrack(spacetrackUser, spacetrackPassword)
    spacetrack.updateTLEsFor(noradId, tleFilePath, latestPhotoTime)

    ephemCalculator = EphemerisCalculator(tleFilePath)    
    cameraPositionsGCRS = map(ephemCalculator, originalPhotoTimes)
    
    wcsHeaders = auromat.solving.solving.solveImages(imagePaths, channel=channel,
                                                     maskingFn=maskingFn,
                                                     solveTimeout=solveTimeout, 
                                                     debugOutputFolder=debugOutputFolder, 
                                                     parallel=parallel, maxWorkers=maxWorkers,
                                                     noAstrometryPlots=noAstrometryPlots,
                                                     sigma=sigma, verbose=verbose,
                                                     oddsToSolve=oddsToSolve)
    
    for (wcsHeader, originalPhotoTime, cameraPosGCRS) in zip(wcsHeaders, originalPhotoTimes, cameraPositionsGCRS):
        if wcsHeader is not None:
            auromat.fits.setNoradId(wcsHeader, noradId)
            auromat.fits.setSpacecraftPosition(wcsHeader, cameraPosGCRS, originalPhotoTime)
        yield wcsHeader

    


# Copyright European Space Agency, 2013

from __future__ import division, absolute_import, print_function

import time
import numpy as np
import numpy.ma as ma

import auromat.utils
from auromat.util.decorators import lazy_property, inherit_docs
from auromat.mapping.mapping import BaseMapping, GenericMapping, inflatedEarthIntersection
from auromat.coordinates.transform import j2000ToLatLon, j2000ToMLatMLT,\
    spherical_to_cartesian
from auromat.utils import vectorLengths
from auromat.coordinates.wcs import pix2world

@inherit_docs
class BaseAstrometryMapping(BaseMapping):
    """
    A mapping which calculates its coordinates based on the camera position and its WCS definition.
    """

    def __init__(self, wcsHeader, alti, cameraPosGCRS, photoTime, identifier, metadata={},
                 fastCenterCalculation=False):
        """
        
        :param alti: mapping altitude in km
        :param fastCenterCalculation:
            Calculates center coordinates directly from the mean of the
            corner coordinates.
        """
        BaseMapping.__init__(self, alti, cameraPosGCRS, photoTime, identifier, metadata)
        self._wcsHeader = wcsHeader
        self.fastCenterCalculation = fastCenterCalculation
        if fastCenterCalculation:
            # When calculating center coordinates from corners it is already assured
            # that the centers are only defined when all its corners are defined.
            # This then carries over to the elevation calculation.
            # Note that img must be masked manually in the subclass!
            self.isSanitized = True
            
        self._mlatmlt = None
        self._mlatmltCenter = None

    @property
    def wcsHeader(self):
        return self._wcsHeader
        
    @lazy_property
    def cameraToPixelCornerDirection(self):
        """
        Direction vector for each pixel corner.
        """
        return pixelDirection(self.wcsHeader, corner=True)
    
    @lazy_property
    def cameraToPixelCenterDirection(self):
        """
        Direction vector for each pixel center.
        """
        if self.fastCenterCalculation:
            return self._calcCenters(self.cameraToPixelCornerDirection)
        else:
            return pixelDirection(self.wcsHeader, corner=False)
        
    @property
    def ra(self):
        """
        Right ascension for each pixel center.
        For debugging purposes only!
        """
        photoWidth, photoHeight = self.wcsHeader['IMAGEW'], self.wcsHeader['IMAGEH']
        ra, _ = pix2world(self.wcsHeader, photoWidth, photoHeight)
        return ra
    
    @property
    def dec(self):
        """
        Descension for each pixel center.
        For debugging purposes only!
        """
        photoWidth, photoHeight = self.wcsHeader['IMAGEW'], self.wcsHeader['IMAGEH']
        _, dec = pix2world(self.wcsHeader, photoWidth, photoHeight)
        return dec
    
    @lazy_property
    def intersectionInflatedCorner(self):
        """
        Returns the point of intersection with the inflated earth for each pixel corner.
        """
        intersectionInflated = inflatedEarthIntersection(self.cameraToPixelCornerDirection.reshape(-1,3),
                                                         self.cameraPosGCRS, self.altitude)
        return intersectionInflated.reshape(self.cameraToPixelCornerDirection.shape)

    @lazy_property
    def intersectionInflatedCenter(self):
        """
        Returns the point of intersection with the inflated earth for each pixel center.
        """
        if self.fastCenterCalculation:
            intersectionInflated = self._calcCenters(self.intersectionInflatedCorner)
        else:
            intersectionInflated = inflatedEarthIntersection(self.cameraToPixelCenterDirection.reshape(-1,3),
                                                             self.cameraPosGCRS, self.altitude)
            intersectionInflated = intersectionInflated.reshape(self.cameraToPixelCenterDirection.shape)
        return intersectionInflated
    
    @property
    def distance(self):
        """
        Distance for each pixel center between camera and intersection point.
        For debugging purposes only!
        """
        vectors = (self.intersectionInflatedCenter - self.cameraPosGCRS).reshape(-1, 3)
        lens = vectorLengths(vectors).reshape(self.intersectionInflatedCenter.shape[0], self.intersectionInflatedCenter.shape[1])
        return lens
    
    @property
    def lats(self):
        lats, _ = self._latsLonsCorner
        return lats
        
    @property
    def lons(self):
        _, lons = self._latsLonsCorner
        return lons
    
    @property
    def latsCenter(self):
        lats, _ = self._latsLonsCenter
        return lats
        
    @property
    def lonsCenter(self):
        _, lons = self._latsLonsCenter
        return lons
        
    @lazy_property
    def _latsLonsCorner(self):
        latDeg, lonDeg = j2000ToLatLon(self.intersectionInflatedCorner.reshape(-1,3), self.photoTime)
        latDeg = latDeg.reshape(self.intersectionInflatedCorner.shape[0], self.intersectionInflatedCorner.shape[1])
        lonDeg = lonDeg.reshape(self.intersectionInflatedCorner.shape[0], self.intersectionInflatedCorner.shape[1])
        latDeg, lonDeg = ma.masked_invalid(latDeg, copy=False), ma.masked_invalid(lonDeg, copy=False)
        return latDeg, lonDeg
    
    @lazy_property
    def _latsLonsCenter(self):
        latDeg, lonDeg = j2000ToLatLon(self.intersectionInflatedCenter.reshape(-1,3), self.photoTime)
        latDeg = latDeg.reshape(self.intersectionInflatedCenter.shape[0], self.intersectionInflatedCenter.shape[1])
        lonDeg = lonDeg.reshape(self.intersectionInflatedCenter.shape[0], self.intersectionInflatedCenter.shape[1])
        latDeg, lonDeg = ma.masked_invalid(latDeg, copy=False), ma.masked_invalid(lonDeg, copy=False)
        return latDeg, lonDeg
    
    @staticmethod
    def _calcCenters(corners):
        centers = corners[:-1,:-1] + corners[:-1,1:]
        centers += corners[1:,1:]
        centers += corners[1:,:-1]
        centers /= 4
        return centers
    
    def setDirty(self):
        """
        Overrides BaseMapping.setDirty()
        """
        self._mlatmlt = None
        self._mlatmltCenter = None        
        super(BaseAstrometryMapping, self).setDirty()
    
    @property
    def mLatMlt(self):
        """
        Overrides BaseMapping.mLatMlt. 
        We directly use the J2000 coordinates (instead of GEO) as source here to minimize
        numerical errors caused by additional transformations and to gain some speed.
        """
        if self._mlatmlt is None:
            mlat, mlt = j2000ToMLatMLT(self.intersectionInflatedCorner.reshape(-1,3), self.photoTime)
            mlat = mlat.reshape(self.intersectionInflatedCorner.shape[0], self.intersectionInflatedCorner.shape[1])
            mlt = mlt.reshape(self.intersectionInflatedCorner.shape[0], self.intersectionInflatedCorner.shape[1])
            mask = ma.getmaskarray(self.lats)
            self._mlatmlt = ma.masked_array(mlat, mask), ma.masked_array(mlt, mask)
        return self._mlatmlt
    
    @property
    def mLatMltCenter(self):
        """
        Overrides BaseMapping.mLatMltCenter.
        We directly use the J2000 coordinates (instead of Lat/Lon->GEO) as source here to minimize
        numerical errors caused by additional transformations and to gain some speed.
        """
        if self._mlatmltCenter is None:
            mlat, mlt = j2000ToMLatMLT(self.intersectionInflatedCenter.reshape(-1,3), self.photoTime)
            mlat = mlat.reshape(self.intersectionInflatedCenter.shape[0], self.intersectionInflatedCenter.shape[1])
            mlt = mlt.reshape(self.intersectionInflatedCenter.shape[0], self.intersectionInflatedCenter.shape[1])
            mask = ma.getmaskarray(self.latsCenter)
            self._mlatmltCenter = ma.masked_array(mlat, mask), ma.masked_array(mlt, mask)
        return self._mlatmltCenter
        
    @lazy_property
    def elevation(self):
        """
        Elevation for each pixel center. Angles are between 0 (horizon) and 90 (nadir) degrees.
        """
        pixelToCameraDirection = -self.cameraToPixelCenterDirection
        intersectionUnit = auromat.utils.unitVectors(self.intersectionInflatedCenter.reshape(-1,3))
        alpha = auromat.utils.angleBetween(pixelToCameraDirection.reshape(-1,3), intersectionUnit)
        alpha = alpha.reshape(self.intersectionInflatedCenter.shape[0], self.intersectionInflatedCenter.shape[1])
        np.rad2deg(alpha, alpha)
        np.subtract(90, alpha, alpha)
        elevation = ma.masked_invalid(alpha, copy=False)
        return elevation
    
    def createResampled(self, lats, lons, latsCenter, lonsCenter, elevation, img):
        mapping = GenericMapping(lats, lons, latsCenter, lonsCenter, elevation, self.altitude, img, 
                                 self.cameraPosGCRS, self.photoTime, self.identifier,
                                 metadata=self.metadata)
        return mapping
 
class ImageMaskAstrometryMixin(object):
    """
    Helper mixin class which handles the fastCenterCalculation=True case for
    images.
    
    Has to be applied as last step in the hierarchy, e.g.:
    
    MyClass(ImageMaskAstrometryMixin, FileImageMixin, BaseSpacecraftMapping)
    """
    
    def __init__(self):
        self._img = None
    
    @property
    def img(self):
        if self._img is None:
            img = super(ImageMaskAstrometryMixin, self).img
            if self.fastCenterCalculation:
                # see BaseAstrometryMapping.__init__ for why this is needed
                mask = ma.getmaskarray(self.latsCenter)
                imgMask = np.repeat(mask[:,:,None], 3, 2)
                img = ma.masked_array(img.data, mask=imgMask)
            self._img = img
        return self._img

def pixelDirection(fitsWcsHeader, corner=True):
    """
    Calculates the direction vector in ICRS for each pixel corner or center, given a WCS solution.
    
    Technically, the returned cartesian ICRS coordinates would have to be converted
    to GCRS/J2000 for the earth-intersection calculations,
    but this would involve distances which aren't available here.
    The solution is that we pretend that the ICRS coordinates are GCRS coordinates. This is acceptable because
    the error is lower than pixel resolution. 
    (ICRF/GCRS diff is around 0.01" while ISS photography is around 20-100"/px)
    
    :param dictionary fitsWcsHeader: must also contain IMAGEW, IMAGEH in pixels
    :rtype: unit direction vector array of shape (IMAGEH+1, IMAGEW+1, 3) if corner==True,
            otherwise (IMAGEH, IMAGEW, 3) 
    """
    photoWidth, photoHeight = fitsWcsHeader['IMAGEW'], fitsWcsHeader['IMAGEH']

    t0 = time.time()   
    camToPixelDirection = pix2world(fitsWcsHeader, photoWidth, photoHeight, corner=corner, ascartesian=True)
    print('pix2world:', time.time()-t0, 's')
    
    assert np.all(np.array(camToPixelDirection.shape) == np.array([photoHeight + corner, photoWidth + corner, 3])), \
           '{} != {}'.format(camToPixelDirection.shape, [photoHeight + corner, photoWidth + corner, 3])
    
    return camToPixelDirection



# Copyright European Space Agency, 2013

from __future__ import print_function

import os.path
import cProfile
from auromat.mapping.spacecraft import getMapping
from auromat.resample import resample

try:
    import pyinstrument
except ImportError:
    print('pyinstrument not available, will use cProfile only')
    pyinstrument = None

def cprofile_profile(func):
    def profiled_func(*args, **kwargs):
        profile = cProfile.Profile()
        try:
            profile.enable()
            result = func(*args, **kwargs)
            profile.disable()
            return result
        finally:
            profile.print_stats(sort='time')
    return profiled_func

def pyinstrument_profile(func):
    def profiled_func(*args, **kwargs):
        try:
            profile = pyinstrument.Profiler()
        except:
            profile = pyinstrument.Profiler(use_signal=False)
        try:
            profile.start()
            result = func(*args, **kwargs)
            profile.stop()
            return result
        finally:
            print(profile.output_text(color=True))
    return profiled_func

def profile(profiler):
    imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
    wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
    m = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    profiler(_profileLatsLons)(m)
    profiler(_profileBoundingBox)(m)
    profiler(_profileMeanResampling)(m)

def _profileLatsLons(m):
    print('profiling lats/lons/elevation/img calculation')
    m.lats
    m.latsCenter
    m.elevation
    m.img

def _profileBoundingBox(m):
    print('profiling bounding box calculation')
    m.boundingBox

def _profileMeanResampling(m):
    print('profiling mean resampling')
    resample(m, method='mean')

def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)

if __name__ == '__main__':
    profile(cprofile_profile)
    
    if pyinstrument is not None:
        profile(pyinstrument_profile)



# Copyright European Space Agency, 2013

"""
Some methods for easy debugging of parameters on the shell.
"""

from __future__ import division

import os
import sys
import tempfile
import subprocess

import auromat.mapping.spacecraft
import auromat.resample
import auromat.draw
from auromat.util import exiftool
from auromat.util.image import saveImage, loadImage
from auromat.solving.masking import maskStarfield
from auromat.draw import saveFig

def debugHorizon(imagePath, fitsWcsPath, tleFolder, photoTimeDelta=None, altitude=110):
    """
    Visualizes the horizon of the modelled earth and inflated earth (=aurora)
    for the given parameters.
    
    :param datetime.timedelta photoTimeDelta: amount to shift the image time (affects calculated
        camera position and therefore the geodetic coordinates)
    """    
    mapping = auromat.mapping.spacecraft.getMapping(imagePath, fitsWcsPath, photoTimeDelta,
                                                   tleFolder=tleFolder, altitude=altitude) 
    
    tmpPath = tempfile.mktemp(suffix='.jpg')
    saveFig(tmpPath, auromat.draw.drawHorizon(mapping))
    
    _openFile(tmpPath)
    
def debugPlot(imagePath, fitsWcsPath, tleFolder, photoTimeDelta=None, altitude=0, minElevation=10):
    """
    Plot the georeferenced image on a map. This allows to play with the image timestamp
    and mapping altitude.
    
    :param imagePath:
    :param fitsWcsPath:
    :param tleFolder:
    :param datetime.timedelta photoTimeDelta: amount to shift the image time (affects calculated
        camera position and therefore the geodetic coordinates)
    :param altitude: in km
    :param minElevation: in degrees
    """
     
    mapping = auromat.mapping.spacecraft.getMapping(imagePath, fitsWcsPath, photoTimeDelta, 
                                                   tleFolder=tleFolder, altitude=altitude)
    
    mapping = mapping.maskedByElevation(minElevation)
    mapping = auromat.resample.resample(mapping)
        
    tmpPath = tempfile.mktemp(suffix='.png')
    saveFig(tmpPath, auromat.draw.drawPlot(mapping))
    _openFile(tmpPath)    

def maskAllInFolder(folderPath, outputFolderPath, ext='jpg', maskFn=maskStarfield, preserveExif=True):
    """
    Masks the starfield of all images in `folderPath` and stores the masked images
    in `outputFolderPath`.
    
    :param str ext: image extension
    :param function maskFn: the masking function to use, 
        by default the standard algorithm from :mod:`auromat.solving.masking`
    :param bool preserveExif: whether to copy EXIF data (exiftool must be installed)
    """
    if preserveExif:
        et = exiftool.ExifTool()
        et.start()
    
    for imageFilename in os.listdir(folderPath):
        imagePath = os.path.join(folderPath, imageFilename)
        maskedPath = os.path.join(outputFolderPath, os.path.splitext(imageFilename)[0] + '.' + ext)
        mask, _ = maskFn(imagePath)
        im = loadImage(imagePath)
        im[~mask] = 0
        saveImage(maskedPath, im)
        
        if preserveExif:
            et.copy_tags(imagePath, maskedPath)
    
    if preserveExif:
        et.terminate()
    
def _openFile(path):
    if sys.platform.startswith('darwin'):
        subprocess.call(('open', path))
    elif os.name == 'nt':
        subprocess.call(('start', path), shell=True)
    elif os.name == 'posix':
        subprocess.call(('xdg-open', path))


# Copyright European Space Agency, 2013

from __future__ import division, print_function

from collections import namedtuple
import datetime
import os
import fnmatch

import numpy as np
import numpy.ma as ma
from numpy.core.umath_tests import matrix_multiply

from astropy.coordinates.angles import Angle
import astropy.units as u

from auromat.coordinates.geodesic import wgs84A, wgs84B
from auromat.mapping.mapping import BaseMapping, BoundingBox, GenericMapping,\
    BaseMappingProvider, MappingCollection, \
    inflatedEarthIntersection, DefaultRGBMixin, sanitize_data
from auromat.coordinates.transform import Y, Z, rotation_matrix, geodetic2EcefZero,\
    ecef2Geodetic, spherical_to_cartesian, latLonToJ2000
import auromat.utils
from auromat.util.decorators import lazy_property, inherit_docs
from auromat.util.image import loadImage

fileDateTimeFormat = '%y%m%d_%H%M%S'

# NOTE: xc, yc, k are relative to a 512x512 image
# NOTE: xc is vertical, yc is horizontal!
CalibrationData = namedtuple('CalibrationData', ['station', 'validFrom', 'validTo',
                                                 'lat', 'lon', 'xc', 'yc', 
                                                 'k', 'rotation',
                                                 'boundingBoxSimple'])

@inherit_docs
class MIRACLEMappingProvider(BaseMappingProvider):
    
    def __init__(self, imageFolder, altitude=110, simple=False, maxTimeOffset=5):
        """
        
        :param imageFolder: path to folder containing images and cal.txt file
        :param altitude: the altitude in km onto which the image was mapped (e.g. 110)
                               Note that for simple=True, this is always 110km
        :param bool simple: whether to use simple mapping (constant lat-lon grid) or
                            intersection-based mapping
        :param maxTimeOffset: in seconds
        
        """
        self.imageFolder = imageFolder
        self.altitude = altitude
        self.simple = simple
        self.imageFileExtension = 'jpg'
        self.maxTimeOffset = maxTimeOffset
        
        fileNames = os.listdir(imageFolder)
        imageFileNames = sorted(fnmatch.filter(fileNames, '*.' + self.imageFileExtension))
        imageStations = [f[:3] for f in imageFileNames]
        self.imageDates = [datetime.datetime.strptime(f[3:16], fileDateTimeFormat) for f in imageFileNames]
        
        self.images = dict() # station -> [(filename,date)]
        for station in set(imageStations):
            self.images[station] = [(f,d) for (f,s,d) in zip(imageFileNames, imageStations, self.imageDates) if s == station]
    
    def __len__(self):
        return len(self.imageDates)
    
    @property
    def range(self):
        dates = sorted(self.imageDates)
        return dates[0], dates[-1]
    
    def contains(self, date):
        for images in self.images.values():
            dates = [d for (_,d) in images]
            
            # search for image at given time
            idx = auromat.utils.findNearest(dates, date)
            offset = abs(dates[idx]-date).total_seconds()
            if offset <= self.maxTimeOffset:
                return True
        
        return False
        
    def get(self, date):
        mappings = []
        
        # for each station find the image closest to date+-maxTimeOffset
        for images in self.images.values():
            dates = [d for (_,d) in images]
            
            # search for image at given time
            idx = auromat.utils.findNearest(dates, date)
            offset = abs(dates[idx]-date).total_seconds()
            if offset <= self.maxTimeOffset:
                imagePath = os.path.join(self.imageFolder, images[idx][0])
                mappings.append(getMapping(imagePath, self.altitude, self.simple))

        ident =  'MIRACLE.' + date.strftime('%Y.%m.%d.%H.%M.%S')            
        mappingColl = MappingCollection(mappings, identifier=ident, mayOverlap=True)
        return mappingColl
        
    def getById(self, identifier):
        raise NotImplementedError
    
    def getSequence(self, dateBegin=None, dateEnd=None):
        raise NotImplementedError

@sanitize_data
@inherit_docs
class MIRACLEMapping(DefaultRGBMixin, BaseMapping):
    """
    A mapping defined using an image and calibration data from FMI MIRACLE.
    """
    
    def __init__(self, calData, imagePath, photoTime, alti, simple=False):
        """
        
        :param CalibrationData calData:
        :param photoTime: datetime object
        :param alti: the altitude in km onto which the image was mapped (e.g. 110)
                     Note that for simple=True, this is always 110km
        :param bool simple: whether to use simple mapping (constant lat-lon grid) or
                            intersection-based mapping
        """
        
        # TODO SOD ASI images include wavelength in filename, add to identifier?
        identifier = calData.station + '.' + photoTime.strftime('%Y.%m.%d.%H.%M.%S')
        
        xASI,yASI,zASI = geodetic2EcefZero(np.deg2rad(calData.lat), np.deg2rad(calData.lon))
        self.cameraPosGEO = [xASI, yASI, zASI]

        cameraPosGCRS = latLonToJ2000(calData.lat, calData.lon, 0, photoTime)
        
        alti = 110 if simple or alti is None else alti
        
        BaseMapping.__init__(self, alti, cameraPosGCRS, photoTime, identifier)
        
        self._calData = calData
        self._simple = simple
        self._imagePath = imagePath
        self._img = None
        self._img_unmasked = None
    
    @property
    def img(self):
        if self._img is None:
            self._img = ma.masked_array(self._img_unmasked)   
        return self._img
    
    @property
    def img_unmasked(self):
        if self._img_unmasked is None:
            rgb = loadImage(self._imagePath)
            if rgb.shape[0] != rgb.shape[1]:
                # contains caption below image, cut it off
                rgb = rgb[:rgb.shape[1],:]
                assert rgb.shape == (rgb.shape[1],rgb.shape[1],3)
            self._img_unmasked = rgb
        return self._img_unmasked
    
    @property
    def lats(self):
        lats, _ = self._latsLonsCorner
        return lats
        
    @property
    def lons(self):
        _, lons = self._latsLonsCorner
        return lons
    
    @property
    def latsCenter(self):
        lats, _ = self._latsLonsCenter
        return lats
        
    @property
    def lonsCenter(self):
        _, lons = self._latsLonsCenter
        return lons
        
    @lazy_property
    def _latsLonsCorner(self):
        lats, lons = self._calculateLatsLons(center=False)
        lats = ma.masked_invalid(lats, copy=False)
        lons = ma.masked_invalid(lons, copy=False)
        return lats, lons
    
    @lazy_property
    def _latsLonsCenter(self):
        lats, lons = self._calculateLatsLons(center=True)
        lats = ma.masked_invalid(lats, copy=False)
        lons = ma.masked_invalid(lons, copy=False)
        return lats, lons
    
    def _calculateLatsLons(self, center=True):
        if self._simple:
            bb = self._calData.boundingBoxSimple
            
            deltaLat = (bb.latNorth - bb.latSouth)/self.img.shape[0]
            deltaLon = (bb.lonEast - bb.lonWest)/self.img.shape[1]
            
            if center:
                latSpace = np.linspace(bb.latNorth-deltaLat/2, bb.latSouth+deltaLat/2, self.img.shape[0])
                lonSpace = np.linspace(bb.lonWest+deltaLon/2, bb.lonEast-deltaLon/2, self.img.shape[0])
            else:
                latSpace = np.linspace(bb.latNorth, bb.latSouth, self.img.shape[0]+1)
                lonSpace = np.linspace(bb.lonWest, bb.lonEast, self.img.shape[0]+1)
                
            #lats, lons = np.meshgrid(latSpace, lonSpace, indexing='ij') # 'indexing' not supported in np 1.6
            lats, lons = np.dstack(np.meshgrid(latSpace, lonSpace)).T
        else:
            intersectionInflated = self.intersectionInflatedCenter if center else self.intersectionInflatedCorner
            xyz = intersectionInflated.reshape(-1,3)
            lats, lons = ecef2Geodetic(xyz[:,0], xyz[:,1], xyz[:,2], wgs84A, wgs84B)
            np.rad2deg(lats, lats)
            np.rad2deg(lons, lons)
            lats = lats.reshape(intersectionInflated.shape[0], intersectionInflated.shape[1])
            lons = lons.reshape(intersectionInflated.shape[0], intersectionInflated.shape[1])
        
        return lats, lons
        
    @lazy_property
    def cameraToPixelCornerDirection(self):
        """
        Direction vector for each pixel corner.
        """        
        vecs = self._calculateCameraToPixelDirection(self.elevationCorner, self.azimuthCorner)
        return vecs
    
    @lazy_property
    def cameraToPixelCenterDirection(self):
        """
        Direction vector for each pixel center.
        """
        _, el = self.azElCenter # don't use self.elevation here, conflicts with sanitization decorator
        vecs = self._calculateCameraToPixelDirection(el, self.azimuthCenter)
        return vecs
    
    def _calculateCameraToPixelDirection(self, el, az):
        el = np.deg2rad(el)
        az = np.deg2rad(-(az-180))
        
        x,y,z = spherical_to_cartesian(1,el,az)

        vecs = np.dstack((x,y,z))
        
        # simple spherical latitude rotation works here because
        # the latitude is the geodetic latitude which is the
        # angle between the normal and the equatorial plane
        matLat = rotation_matrix(np.deg2rad(90 - self._calData.lat), Y)[:3,:3]
        matLon = rotation_matrix(np.deg2rad(-self._calData.lon), Z)[:3,:3]
        mat = np.dot(matLon, matLat) # rotate latitude first, then longitude
        
        vecs = vecs.reshape(el.shape[0]*el.shape[1], 3)
        vecsRot = matrix_multiply(mat, vecs[...,np.newaxis]).reshape(el.shape[0], el.shape[1], 3)
                
        return vecsRot
                
    @lazy_property
    def intersectionInflatedCorner(self):
        """Returns the point of intersection with the inflated earth for each pixel corner."""        
        intersectionInflated = inflatedEarthIntersection(self.cameraToPixelCornerDirection.reshape(-1,3),
                                                         self.cameraPosGEO, self.altitude)
        return intersectionInflated.reshape(self.cameraToPixelCornerDirection.shape)
    
    @lazy_property
    def intersectionInflatedCenter(self):
        """Returns the point of intersection with the inflated earth for each pixel center."""
        intersectionInflated = inflatedEarthIntersection(self.cameraToPixelCenterDirection.reshape(-1,3),
                                                         self.cameraPosGEO, self.altitude)
        return intersectionInflated.reshape(self.cameraToPixelCenterDirection.shape)

    @property
    def elevation(self):
        _, el = self.azElCenter
        return el
    
    @property
    def elevationCorner(self):
        _, el = self.azElCorner
        return el
    
    @property
    def azimuthCenter(self):
        az, _ = self.azElCenter
        return az

    @property
    def azimuthCorner(self):
        az, _ = self.azElCorner
        return az
    
    @lazy_property
    def azElCorner(self):
        """
        Azimuth and elevation for each pixel corner.
        """
        az, el = self.calculateAzEl(center=False)
        az = ma.masked_invalid(az, copy=False)
        el = ma.masked_invalid(el, copy=False)
        return az, el
    
    @lazy_property
    def azElCenter(self):
        """
        Azimuth and elevation for each pixel center.
        """
        az, el = self.calculateAzEl(center=True)
        az = ma.masked_invalid(az, copy=False)
        el = ma.masked_invalid(el, copy=False)
        return az, el
    
    def calculateAzEl(self, center=True):
        w = self.img_unmasked.shape[0]
        xc = self._calData.xc
        yc = self._calData.yc
        k = self._calData.k
        
        refW = 512
        if w != refW:
            xc *= w/refW
            yc *= w/refW
            k *= w/refW
        
        w_ = w if center else w+1
        
        center_ = np.array([xc,yc])
        ind = np.indices((w_, w_))
        # TODO verify that the (0,0) is really the upper left corner of the upper left pixel
        if center:
            ind += 0.5
        ind = np.dstack((ind[0], ind[1])).reshape((w_)*(w_), 2)
        vecs = ind - center_
        
        northVec = [-1,0] # points to top of image
        northVecs = np.repeat([northVec], len(vecs), axis=0)        
        az = auromat.utils.signedAngleBetween(vecs, northVecs).reshape(w_,w_)
        az -= self._calData.rotation
        az = Angle(az * u.rad).wrap_at(360 * u.deg).degree
        
        distCenter = np.sqrt((vecs*vecs).sum(axis=1)).reshape(w_,w_)
        z = distCenter/k
        np.rad2deg(z, z)
        elev = 90 - z
        
        return az, elev
               
    def createResampled(self, lats, lons, latsCenter, lonsCenter, elevation, img):
        mapping = GenericMapping(lats, lons, latsCenter, lonsCenter, elevation, self.altitude, img, 
                                 self.cameraPosGCRS, self.photoTime, self.identifier)
        return mapping
    
def getMapping(imagePath, alti=110, simple=False):
    filename = os.path.basename(imagePath)
    # e.g. KEV120304_172100_____4000.jpg
    station = filename[:3]
    date = datetime.datetime.strptime(filename[3:16], fileDateTimeFormat)
    
    calPath = os.path.join(os.path.dirname(imagePath), 'cal.txt')
    calData = getCalibrationData(calPath, station, date)
    
    mapping = MIRACLEMapping(calData, imagePath, date, alti, simple=simple)
    mapping = mapping.maskedByElevation(0.1) # .1 to account for rounding errors
    return mapping

def getCalibrationData(path, station, date):
    
    calDataEntries = np.loadtxt(path, dtype={'names': ('station', 'lat', 'lon', 'from', 'to', 
                                                       'xc', 'yc', 'k', 'rotation', 'lat+', 'lat-', 
                                                       'lon-', 'lon+', 'i1', 'i2', 'i3'),
                                             'formats': ('S3', 'f8', 'f8', 'f8', 'f8', 
                                                         'f8', 'f8', 'f8', 'f8', 'f8', 'f8', 
                                                         'f8', 'f8', 'b1', 'b1', 'b1')})
    for entry in calDataEntries:
        if entry['station'] != station:
            continue
        
        fromDateY = int(entry['from'])
        fromDateM = int((entry['from']-fromDateY)*12 + 1)
        toDateY = int(entry['to'])
        toDateM = int((entry['to']-toDateY)*12 + 1)
        
        fromDate = datetime.datetime(fromDateY, fromDateM, 1)
        toDate = datetime.datetime(toDateY, toDateM+1, 1) # easier than using end of month
        
        if not fromDate <= date <= toDate:
            continue
        
        lat = entry['lat']
        lon = entry['lon']
        
        bbSimple = BoundingBox(latSouth=lat+entry['lat-'], 
                               lonWest=lon+entry['lon-'], 
                               latNorth=lat+entry['lat+'], 
                               lonEast=lon+entry['lon+'])
        
        calData = CalibrationData(station=entry['station'], validFrom=fromDate, validTo=toDate, 
                                  lat=lat, lon=lon, 
                                  xc=entry['xc'], yc=entry['yc'], k=entry['k'],
                                  rotation=entry['rotation'], boundingBoxSimple=bbSimple)
        return calData
        
    raise ValueError('No MIRACLE calibration data found for ' + station + ' station')

# TODO MIRACLEMapping is missing in sphinx docs
__all__ = map(lambda f: f.__name__,
              [MIRACLEMappingProvider, MIRACLEMapping, getMapping])


# Copyright European Space Agency, 2013

from __future__ import absolute_import, division

import astropy.time

def containsLeapSecond(d1, d2):
    """
    Checks whether the given datetime range contains a leap second.
    
    :param datetime.datetime d1: start date
    :param datetime.datetime d2: end date
    :rtype: bool
    """
    eps = 1e-6
    diffWithoutLeap = (d2-d1).total_seconds()
    diffWithLeap = (astropy.time.Time(d2, scale='utc') - astropy.time.Time(d1, scale='utc')).sec
    containsLeap = abs(diffWithoutLeap - diffWithLeap) > eps
    return containsLeap
        


# Copyright European Space Agency, 2013

from __future__ import print_function

import timeit

setup = """
import numpy as np
from astropy.coordinates.distances import spherical_to_cartesian, cartesian_to_spherical
from auromat.coordinates.transform import _ecef2GeodeticOptimized_np, _ecef2GeodeticOptimized_ne
from auromat.coordinates.transform import _spherical_to_cartesian_np, _spherical_to_cartesian_ne
from auromat.coordinates.transform import _cartesian_to_spherical_np, _cartesian_to_spherical_ne

N = 10*1000*1000
x = np.random.rand(N)
y = np.random.rand(N)
z = np.random.rand(N)
"""

setup2 = setup + """
from auromat.coordinates.intersection import _ellipsoidLineIntersection_np, _ellipsoidLineIntersection_ne
from auromat.coordinates.intersection import _ellipsoidLineIntersects_np, _ellipsoidLineIntersects_ne

lineDirection = np.array([x,y,z]).T
a = b = 2
"""

if __name__ == "__main__":
    
    print(min(timeit.repeat('_ecef2GeodeticOptimized_np(x, y, z)', setup, number=1)), 'vs.',\
          min(timeit.repeat('_ecef2GeodeticOptimized_ne(x, y, z)', setup, number=1)))
    
#    print min(timeit.repeat('spherical_to_cartesian(x, y, z)', setup, number=1)), 'vs.',\
#          min(timeit.repeat('_spherical_to_cartesian_np(x, y, z)', setup, number=1)), 'vs.',\
#          min(timeit.repeat('_spherical_to_cartesian_ne(x, y, z)', setup, number=1))

#    print min(timeit.repeat('cartesian_to_spherical(x, y, z)', setup, number=1)), 'vs.',\
#          min(timeit.repeat('_cartesian_to_spherical_np(x, y, z)', setup, number=1)), 'vs.',\
#          min(timeit.repeat('_cartesian_to_spherical_ne(x, y, z)', setup, number=1))
          
#    print min(timeit.repeat('_ellipsoidLineIntersection_np(a, b, [0,0,0], lineDirection)', setup2, number=1)), 'vs.',\
#          min(timeit.repeat('_ellipsoidLineIntersection_ne(a, b, [0,0,0], lineDirection)', setup2, number=1))
          
#    print min(timeit.repeat('_ellipsoidLineIntersects_np(a, b, [0,0,0], lineDirection)', setup2, number=1)), 'vs.',\
#          min(timeit.repeat('_ellipsoidLineIntersects_ne(a, b, [0,0,0], lineDirection)', setup2, number=1))


# Copyright European Space Agency, 2013

# The following code is commented out because nansat cannot currently be used 
# due to incompatible licensing. Once auromat is released under a real open-source
# license, this code should be uncommented again.


"""
Converts mappings to the GDAL data model such that they can be saved in various
raster geospatial file formats such as GeoTIFF, or running data conversions using
gdal tools like gdalwarp.
See http://www.gdal.org/formats_list.html for a list of supported formats.

GDAL's geolocation arrays are our irregular lat/lon arrays.
GDAL supports reading those from netCDF-CF files, but not writing them.
This is one of the reasons why we have a separate netCDF exporter.

Instead of using this module, the gdal tools (like gdalwarp) can also be used
directly on the exported netCDF-CF files. The advantage is that it may integrate
better into a shell based processing pipeline. The disadvantage is that it requires
saving the mappings as (quite huge) netCDF-CF files first.

NOTE: GDAL support is limited currently
      see http://trac.osgeo.org/gdal/ticket/4513#comment:23
"""

#from __future__ import division
#
#import numpy as np
#import nansat
#from osgeo.gdalconst import GCI_RedBand, GCI_GreenBand, GCI_BlueBand
#
## we use Nansat as a nice wrapper around GDAL
#
#def asGdalDataset(mapping):
#    n = asNansat(mapping)
#    dataset = n.vrt.dataset
#    return dataset
#
#def asNansat(mapping):
#    domain = nansat.Domain(lon=mapping.lonsCenter, lat=mapping.latsCenter)
#
#    elevParams = {'name': 'elevation',
#                  }
#    
#    n = nansat.Nansat(domain=domain, array=mapping.elevation, parameters=elevParams)
#    
#    # The min/max data values cannot be set programmatically in GDAL.
#    # The min/max of the data type is used here.
#    # Therefore we leave the img data type as it is and provide the mask
#    # as a separate band.
#    
#    # 0 = no data (opposite to numpy definition)
#    mask = (~mapping.img.mask).astype(np.int8)
#    mask[mask==1] = 64 # see nansat.mosaic
#    # this is not a real GDAL mask band (CreateMaskBand) but is recognized by nansat
#    n.addBand(mask, parameters={'name': 'mask'})
#    
#    #n.vrt.set_subsetMask(self, maskDs, xOff, yOff, dstXSize, dstYSize)
#    
#    if mapping.img.shape[2] == 3: # RGB
#        n.addBand(mapping.img[:,:,0].data, parameters={'name': 'img_r'})
#        n.addBand(mapping.img[:,:,1].data, parameters={'name': 'img_g'})
#        n.addBand(mapping.img[:,:,2].data, parameters={'name': 'img_b'})
#        
#        r = n.get_GDALRasterBand('img_r')
#        r.SetRasterColorInterpretation(GCI_RedBand)
#        
#        g = n.get_GDALRasterBand('img_g')
#        g.SetRasterColorInterpretation(GCI_GreenBand)
#        
#        b = n.get_GDALRasterBand('img_b')
#        b.SetRasterColorInterpretation(GCI_BlueBand)           
#        
#    else: # something else
#        for i in range(mapping.img.shape[2]):
#            n.addBand(mapping.img[:,:,i].data, parameters={'name': 'img'})
#            
#    return n



"""
This package contains command-line tools that will be installed as `auromat-name`
for a module name `auromat.cli.name`. It is not intended to be used from within
Python code. The existing command-line tool modules may be used as references
to develop new special-purpose command-line tools.
"""


# Copyright European Space Agency, 2013

from __future__ import division, absolute_import, print_function

from math import floor, ceil

__all__ = []

NUM_IGRF_YEARS_DEFINED = 24

IGRF_DEFINED_UNTIL_YEAR = 1900 + (NUM_IGRF_YEARS_DEFINED-1)*5

__doc__ = """
This module defines the first three IGRF coefficients `g01`, `g11`, and `h11`
for the years 1900 until {}. They are used by the :mod:`auromat.coordinates.transform`
module.

Note that the coefficients have to be updated here if needed.
When updating, the variables `NUM_IGRF_YEARS_DEFINED`, `g01`, `g11`, and `h11`
have to be changed.
""".format(IGRF_DEFINED_UNTIL_YEAR)

g01 = [-31543, -31464, -31354, -31212, -31060, -30926, -30805, -30715,
       -30654, -30594, -30554, -30500, -30421, -30334, -30220, -30100,
       -29992, -29873, -29775, -29692, -29619.4, -29554.63, -29496.5,
       -29439.5]

g11 = [-2298, -2298, -2297, -2306, -2317, -2318, -2316, -2306, -2292, -2285,
       -2250, -2215, -2169, -2119, -2068, -2013, -1956, -1905, -1848, -1784,
       -1728.2, -1669.05, -1585.9, -1502.4]

h11 = [5922, 5909, 5898, 5875, 5845, 5817, 5808, 5812, 5821, 5810, 5815,
       5820, 5791, 5776, 5737, 5675, 5604, 5500, 5406, 5306, 5186.1, 5077.99, 
       4945.1, 4801.1]

assert len(g01) == len(g11) == len(h11) == NUM_IGRF_YEARS_DEFINED

def calcG01(fracYearIndex, fracYear):
    _checkFracYearIndex(fracYearIndex)
    return (g01[int(floor(fracYearIndex))]*(1.0-fracYear) + 
            g01[int(ceil(fracYearIndex))]*fracYear)

def calcG11(fracYearIndex, fracYear):
    _checkFracYearIndex(fracYearIndex)
    return (g11[int(floor(fracYearIndex))]*(1.0-fracYear) + 
            g11[int(ceil(fracYearIndex))]*fracYear)

def calcH11(fracYearIndex, fracYear):        
    _checkFracYearIndex(fracYearIndex)
    return (h11[int(floor(fracYearIndex))]*(1.0-fracYear) + 
            h11[int(ceil(fracYearIndex))]*fracYear)
    
def _checkFracYearIndex(fracYearIndex):
    if fracYearIndex >= NUM_IGRF_YEARS_DEFINED-1:
        raise ValueError("ERROR: Specified year is greater than IGRF implementation, "
                         "please update coefficients in auromat.coordinates.igrf module")
    


# Copyright European Space Agency, 2013

from __future__ import division, absolute_import

import unittest
from nose.plugins.attrib import attr
from numpy.ma.testutils import assert_array_almost_equal,\
    assert_array_equal, assert_equal
import os
import tempfile
from netCDF4 import Dataset

from auromat.mapping.spacecraft import getMapping
from auromat.resample import resample
import auromat.export.netcdf
from auromat.mapping.netcdf import NetCDFMapping

@attr('slow')
class Test(unittest.TestCase):

    def testRawNetCDFExport(self):
        imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
        wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
        mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
        mapping.checkGuarantees()
        path = tempfile.mktemp(suffix='.nc')
        try:            
            auromat.export.netcdf.write(path, mapping)
            
            with Dataset(path, 'r') as root:
                var = root.variables
                assert set(var.keys()) == set(['time', 'lat', 'lon', 'altitude', 
                                        'lat_bounds', 'lon_bounds', 'mlat', 
                                        'mlt', 'mlat_bounds', 'mlt_bounds', 
                                        'mcrs', 'img_red', 'img_green', 
                                        'img_blue', 'zenith_angle', 
                                        'camera_pos', 'crs'])
            
            cdfmapping = NetCDFMapping(path)
            cdfmapping.checkGuarantees()
            check_equal(cdfmapping, mapping)
                
        finally:
            if os.path.exists(path):
                os.remove(path)

    def testPlateCarreeNetCDFExport(self):
        imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
        wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
        mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
        mapping = resample(mapping, arcsecPerPx=200)
        path = tempfile.mktemp(suffix='.nc')
        try:
            auromat.export.netcdf.write(path, mapping)
            
            cdfmapping = NetCDFMapping(path)
            cdfmapping.checkGuarantees()
            cdfmapping.checkPlateCarree()
            
            check_equal(cdfmapping, mapping)
        finally:
            if os.path.exists(path):
                os.remove(path)
                
def check_equal(cdfmapping, mapping):
    assert_array_equal(cdfmapping.img.shape, mapping.img.shape)
    assert_array_equal(cdfmapping.lats.shape, mapping.lats.shape)

    assert_array_equal(cdfmapping.img, mapping.img)
    assert_array_equal(cdfmapping.lats, mapping.lats)
    assert_array_equal(cdfmapping.lons, mapping.lons)
    assert_array_equal(cdfmapping.latsCenter, mapping.latsCenter)
    assert_array_equal(cdfmapping.lonsCenter, mapping.lonsCenter)
    
    assert_equal(cdfmapping.boundingBox, mapping.boundingBox)
    
    # elevation is stored as float32, so there is some loss of accuracy
    assert_array_almost_equal(cdfmapping.elevation, mapping.elevation, decimal=5)
    
    assert_equal(cdfmapping.photoTime, mapping.photoTime)
    assert_equal(cdfmapping.cameraPosGCRS, mapping.cameraPosGCRS)
              
                               
def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)



# Copyright European Space Agency, 2013

from __future__ import absolute_import, division

import warnings
import os
from datetime import datetime

import numpy as np
import numpy.ma as ma
import skimage.io
import skimage.color
import exifread

from PIL import Image

def loadImage(imagePath):
    """
    Return RGB image in native color range (e.g. [0,255] for uint8).
    Ignores the alpha channel.
    
    Note that this function is not meant for reading RAW files. 
    Use `rawpy <https://pypi.python.org/pypi/rawpy>`_ with desired postprocessing settings instead.
    
    :param imagePath:
    :rtype: rgb array of shape (height,width,3)
    """
    rgb = skimage.io.imread(imagePath)
    
    if rgb.ndim == 2:
        rgb = skimage.color.gray2rgb(rgb)
    
    # ignore alpha if available
    rgb = rgb[:,:,:3]
            
    assert rgb.ndim == 3 and rgb.shape[2] == 3,\
           imagePath + '; wrong shape: ' + str(rgb.shape)
        
    return rgb

def saveImage(imagePath, im, **kw):
    """
    :param imagePath: output path
    :param im: the image
    
    Additional optional keywords for JPEG:
    
    :param quality: 1 to 95
    :param subsampling: '4:4:4', '4:2:2', or '4:1:1'
    """
    if os.path.splitext(imagePath)[1].lower() in ['.jpeg', '.jpg']:
        # for jpg we use Pillow directly, as skimage curently doesn't allow
        # to set the JPEG quality
        im = Image.fromarray(im)
        im.save(imagePath, optimize=True, **kw)
    else:
        skimage.io.imsave(imagePath, im)

def croppedImage(im, divisible_by=16):
    """
    Return image cropped to the next smaller width/height divisable by the given factor.
    Cropping is done on each side and will fail if the amount to crop cannot
    be divided by 2.
    """
    cropped_height = int(im.shape[0]/divisible_by)*divisible_by
    cropped_width = int(im.shape[1]/divisible_by)*divisible_by
    assert (im.shape[0]-cropped_height) % 2 == 0
    assert (im.shape[1]-cropped_width) % 2 == 0
    vcrop = (im.shape[0]-cropped_height)//2
    hcrop = (im.shape[1]-cropped_width)//2
    im = im[vcrop:vcrop+cropped_height,hcrop:hcrop+cropped_width]
    return im

def _normalizeImage(rgb):
    """
    Converts (masked) image to float [0,255] color range.
    If it is a masked image, then it is converted to a non-masked
    image with invalid entries replaced by nan.
    """    
    if rgb.dtype == np.uint8:
        pass
    
    elif rgb.dtype == np.uint16:
        rgb = rgb * (255/65535)
        
    else:
        raise NotImplementedError('Image format ' + str(rgb.dtype) + ' not supported')
    
    rgb = np.require(rgb, np.float)
    
    if ma.isMaskedArray(rgb):
        rgb = rgb.filled(np.nan)
    
    return rgb

def image2mpl(rgb):
    """
    Converts image to [0,1] range suitable for matplotlib.
    """
    rgb = _normalizeImage(rgb)
    return rgb / 255

def image2cv(rgb):
    """
    Converts RGB image to BGR uint8 image suitable for use within OpenCV.
    """
    import cv2 as cv
    rgb = _normalizeImage(rgb)
    rgb = np.require(rgb, np.uint8)
    bgr = cv.cvtColor(rgb, cv.COLOR_RGB2BGR)
    return bgr

def readExifTime(imagePath):
    """
    Reads the date/time an image was taken from its EXIF header.
    If available, sub second time is used as the fractional part
    of the seconds.
    
    :type imagePath: string
    :rtype: datetime
    """
    with open(imagePath, 'rb') as p:
        tags = exifread.process_file(p, details=False)
        timeTakenExif = tags['EXIF DateTimeOriginal']
        subsec = tags.get('EXIF SubSecTimeOriginal', '0')
    return convertExifDate(str(timeTakenExif), str(subsec))

def convertExifDate(dateStr, subsecStr = None):
    """
    Converts an EXIF date/time string to a datetime object.
    
    :param dateStr: date of format %Y:%m:%d %H:%M:%S
    :param subsecStr: float part of seconds as string, e.g. "05"
    :rtype: datetime.datetime
    """
    if subsecStr is None:
        subsecStr = '0'
    else:
        # If there are no leading zeros, then exiftool returns an int
        # instead of a string. To support this case, we convert whatever
        # we get to a string.
        subsecStr = str(subsecStr)
    return datetime.strptime(dateStr + '.' + subsecStr, '%Y:%m:%d %H:%M:%S.%f')

def readFocalLength35mm(imagePath):
    """
    Reads the focal length corresponding to 35mm film from the EXIF header of an image.
    
    :type imagePath: string
    :rtype: int|None
    """
    with open(imagePath, 'rb') as p:
        tags = exifread.process_file(p, details=False, stop_tag='FocalLengthIn35mmFilm')
        try:
            # Note: FocalLength is a ratio while FocalLengthIn35mmFilm is an integer
            focalLength35mm = tags['EXIF FocalLengthIn35mmFilm'].values[0]
        except KeyError:
            return None
    return focalLength35mm

def readExposureTime(imagePath):
    """
    Reads the exposure time in seconds from the EXIF header of an image.
    
    :type imagePath: string
    :rtype: float|None
    """
    with open(imagePath, 'rb') as p:
        tags = exifread.process_file(p, details=False, stop_tag='ExposureTime')
        try:
            exposureTimeRatio = tags['EXIF ExposureTime'].values[0]
            exposureTime = exposureTimeRatio.num / exposureTimeRatio.den
        except KeyError:
            return None
    return exposureTime



# Copyright European Space Agency, 2013

from __future__ import division, print_function

import numpy as np

import unittest
from nose.plugins.attrib import attr

import os
from auromat.mapping.spacecraft import getMapping

@attr('slow')
class Test(unittest.TestCase):            
    def testReal(self):
        m = _getMapping1()
        print(np.min(m.elevation), np.max(m.elevation))
        assert 0 <= np.min(m.elevation) <= np.max(m.elevation) <= 90
        
        m = _getMapping2()
        print(np.min(m.elevation), np.max(m.elevation))
        assert 0 <= np.min(m.elevation) <= np.max(m.elevation) <= 90

def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)

def _getMapping1():
    imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
    wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
    mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    return mapping

def _getMapping2():
    imagePath = getResourcePath('ISS029-E-8492.jpg')
    wcsPath = getResourcePath('ISS029-E-8492.wcs')
    mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    return mapping




# Copyright European Space Agency, 2013

from __future__ import division, print_function

from auromat.coordinates.igrf import IGRF_DEFINED_UNTIL_YEAR

__doc__ = """
This module contains fast and memory-efficient algorithms to convert
coordinates between different reference frames or representations.

Note that some functions depend on the IGRF model whose parameters are
defined in the :mod:`igrf` module. In this version, parameters are defined
until the year {}.
""".format(IGRF_DEFINED_UNTIL_YEAR)

import time

from math import sin, atan2, cos, sqrt, pi, fmod, atan
import numpy as np
from numpy.core.umath_tests import matrix_multiply

try:
    from numexpr import evaluate as ne
except ImportError:
    ne = None

from astropy.time import Time
from astropy.coordinates import Angle
import astropy.units as u

from auromat.coordinates.geodesic import wgs84A, wgs84B, Location
from auromat.coordinates.igrf import calcG11, calcH11, calcG01
    
import auromat.coordinates.transformations

rotation_matrix = auromat.coordinates.transformations.rotation_matrix

def _spherical_to_cartesian_np(r, lat, lon, astuple=True):
    """
    As astropy.coordinates.distances.spherical_to_cartesian but more
    optimized. lat and lon must be arrays.
    """
    # using (3,..) instead of (...,3) as shape has better memory access performance
    # x, y, and z are then each a contiguous block of memory
    res = np.empty((3,) + lat.shape, lat.dtype)
    x = res[0]
    y = res[1]
    z = res[2]
    np.cos(lat, x)
    if r is not None:
        x *= r
    y[:] = x
    x *= np.cos(lon)
    y *= np.sin(lon)
    np.sin(lat, z)
    if r is not None:
        z *= r
        
    if astuple:
        return x,y,z
    else:
        res = np.rollaxis(res, 0, res.ndim)
        return res

def _spherical_to_cartesian_ne(r, lat, lon, astuple=True):
    # using (3,..) instead of (...,3) as shape has better memory access performance
    # x, y, and z are then each a contiguous block of memory
    res = np.empty((3,) + lat.shape, lat.dtype)
    x = res[0]
    y = res[1]
    z = res[2]
    if r is None:
        ne('cos(lat)', out=x)
    else:
        ne('r * cos(lat)', out=x)
    y[:] = x
    ne('x * cos(lon)', out=x)
    ne('y * sin(lon)', out=y)
    if r is None:
        ne('sin(lat)', out=z)
    else:
        ne('r * sin(lat)', out=z)
    if astuple:
        return x,y,z
    else:
        res = np.rollaxis(res, 0, res.ndim)
        return res

def spherical_to_cartesian(r, lat, lon, astuple=True):
    """
    Convert spherical to cartesian coordinates. Inputs must be arrays.
    
    Equivalent to `astropy.coordinates.distances.spherical_to_cartesian`
    for array inputs but uses less memory and is faster.
    
    :type r: ndarray or None (=1)
    :rtype: tuple (x,y,z) of ndarray's with shape as input
    """
    if ne:
        return _spherical_to_cartesian_ne(r, lat, lon, astuple)
    else:
        return _spherical_to_cartesian_np(r, lat, lon, astuple)

def _cartesian_to_spherical_np(x, y, z, with_radius=True):
    xsq = x*x
    ysq = y*y
    zsq = z*z

    if with_radius:
        r = xsq + ysq
        r += zsq
        np.sqrt(r, r)

    s = xsq
    s += ysq
    np.sqrt(s, s)

    lon = ysq
    np.arctan2(y, x, lon)
    
    lat = zsq
    np.arctan2(z, s, lat)

    if with_radius:
        return r, lat, lon
    else:
        return lat, lon

def _cartesian_to_spherical_ne(x, y, z, with_radius=True):
    if with_radius:
        xy = ne('x*x + y*y')
        r = ne('sqrt(xy + z*z)')
        lat = ne('arctan2(z, sqrt(xy))')
        lon = xy
        ne('arctan2(y, x)', out=lon)
        return r, lat, lon
    else:
        lat = ne('arctan2(z, sqrt(x*x + y*y))')
        lon = ne('arctan2(y, x)')
        return lat, lon

def cartesian_to_spherical(x, y, z, with_radius=True):
    """
    Convert cartesian to spherical coordinates. Inputs must be arrays.
    
    Equivalent to `astropy.coordinates.distances.cartesian_to_spherical`
    for array inputs but uses less memory and is faster.
    
    :rtype: tuple (r,lat,lon) or (lat,lon) of ndarray's with shape as input
    """
    if ne:
        return _cartesian_to_spherical_ne(x, y, z, with_radius)
    else:
        return _cartesian_to_spherical_np(x, y, z, with_radius)

def geodetic2Ecef(lat, lon, h, a=wgs84A, b=wgs84B):
    """
    Converts geodetic to Earth Centered, Earth Fixed coordinates.
    
    Parameters h, a, and b must be given in the same unit. 
    The values of the return tuple then also have this unit.
    
    :param lat: latitude(s) in radians
    :param lon: longitude(s) in radians
    :param h: height(s)
    :param a: equatorial axis of the ellipsoid of revolution
    :param b: polar axis of the ellipsoid of revolution
    :rtype: tuple (x,y,z)
    """
    lat, lon, h = np.asarray(lat), np.asarray(lon), np.asarray(h)
    e2 = (a*a - b*b) / (a*a) # first eccentricity squared
    n = a / np.sqrt(1 - e2*np.sin(lat)**2)
    latCos = np.cos(lat)
    nh = n+h
    x = nh*latCos*np.cos(lon)
    y = nh*latCos*np.sin(lon)
    z = (n*(1-e2)+h)*np.sin(lat)    
    return x,y,z

def geodetic2EcefZero(lat, lon, a=wgs84A, b=wgs84B):
    """
    Fast version of :func:`geodetic2Ecef` for `h=0`.
    
    :param lat: latitude(s) in radians
    :param lon: longitude(s) in radians
    :param a: equatorial axis of the ellipsoid of revolution
    :param b: polar axis of the ellipsoid of revolution
    :rtype: tuple (x,y,z)
    """
    lat, lon = np.asarray(lat), np.asarray(lon)
    e2 = (a*a - b*b) / (a*a) # first eccentricity squared
    n = a / np.sqrt(1 - e2*np.sin(lat)**2)
    latn = n*np.cos(lat)
    x = latn*np.cos(lon)
    y = latn*np.sin(lon)
    z = n*(1-e2)*np.sin(lat)    
    return x,y,z

def ecef2Geodetic(x, y, z, a=wgs84A, b=wgs84B):
    """
    Convert ECEF to geodetic coordinates.
    
    This function uses the Bowring algorithm from 1985.
    
    The accuracy is at least 11 decimals (in degrees).
    
    :param x,y,z: ECEF coordinates
    :param a: equatorial axis of the ellipsoid of revolution
    :param b: polar axis of the ellipsoid of revolution
    :rtype: tuple (lat,lon) in radians
    """
    x, y, z = np.asarray(x), np.asarray(y), np.asarray(z)
    if x.ndim > 0:
        return _ecef2GeodeticOptimized(x, y, z, a, b)
    e2 = (a*a - b*b) / (a*a) # first eccentricity squared
    d = (a*a - b*b) / b
    
    p2 = np.square(x) + np.square(y)
    p = np.sqrt(p2)
    r = np.sqrt(p2 + z*z)
    tu = b*z*(1 + d/r)/(a*p)
    tu2 = tu*tu
    cu3 = (1/np.sqrt(1 + tu2))**3
    su3 = cu3*tu2*tu
    tp = (z + d*su3)/(p - e2*a*cu3)
    lat = np.arctan(tp)
    
    lon = np.arctan2(y,x)
    
    return lat, lon

def _ecef2GeodeticOptimized_ne(x, y, z, a=wgs84A, b=wgs84B):
    x, y, z = np.asarray(x), np.asarray(y), np.asarray(z)

    e2 = (a*a - b*b) / (a*a) # first eccentricity squared 
    d = (a*a - b*b) / b  
    
    p2 = ne('x*x + y*y')
    p = ne('sqrt(p2)')
    tu = p2
    ne('b*z*(1 + d/sqrt(p2 + z*z))/(a*p)', out=tu)
    tu2 = ne('tu*tu')
    cu3 = ne('(1/sqrt(1 + tu2))**3')
    lat = p2
    ne('arctan((z + d*cu3*tu2*tu)/(p - e2*a*cu3))', out=lat)
    
    lon = p
    ne('arctan2(y,x)', out=lon)
    
    return lat, lon

def _ecef2GeodeticOptimized_np(x, y, z, a=wgs84A, b=wgs84B):
    """ array-only, memory-efficient version of ecef2Geodetic, around 10-30% faster """
    e2 = (a*a - b*b) / (a*a)
    d = (a*a - b*b) / b
    
    p2 = np.square(x)
    y2 = np.square(y)
    p2 += y2
    p = y2
    np.sqrt(p2, p)
    r = p2
    z2 = z*z
    r += z2
    np.sqrt(r, r)
    tu = z2
    np.divide(d, r, tu)
    tu += 1
    tu *= b
    tu *= z
    ap = a*p
    tu /= ap
    np.square(tu, r)
    tu2 = r
    cu3 = ap
    np.add(1, tu2, cu3)
    np.sqrt(cu3, cu3)
    np.reciprocal(cu3, cu3)
    cu3 **= 3 # don't replace by the faster cu3*cu3*cu3, it's less accurate!
    su3 = tu
    su3 *= cu3
    su3 *= tu2
    tp = tu2
    np.multiply(d, su3, tp)
    tp += z
    pm = p
    cu3m = cu3
    cu3m *= e2*a
    pm -= cu3m
    tp /= pm
    np.arctan(tp, tp)
    lat = tp
    
    lon = cu3
    np.arctan2(y,x, lon)
    
    return lat, lon

_ecef2GeodeticOptimized = _ecef2GeodeticOptimized_ne if ne else _ecef2GeodeticOptimized_np

def rotatePole(lats, lons, altitude, angle=90, axis=[1,0,0]):
    """
    Rotates the given geodetic lat/lon coordinates around the origin.
    
    :param lats, lons: shape (n,) in radians
    :param altitude: in km
    :param angle: degrees
    :param axis: [1, 0, 0], [0, 1, 0], or [0, 0, 1] for x y z axis
    :rtype: tuple (lats, lons) in radians
    """
    assert lats.ndim == 1 and lons.ndim == 1
    assert len(axis) == 3    

    x,y,z = geodetic2Ecef(lats, lons, altitude, wgs84A, wgs84B)
    xyz = np.asarray([x,y,z]).T
    
    alpha = np.deg2rad(angle)
    rot = rotation_matrix(alpha, axis)[:3,:3]
    
    xyzRot = matrix_multiply(rot,xyz[...,np.newaxis]).reshape(xyz.shape)
    lats, lons = ecef2Geodetic(xyzRot[:,0], xyzRot[:,1], xyzRot[:,2], wgs84A, wgs84B)
    return lats, lons

def j2000ToLatLon(j2000Vecs, time_):
    """
    Convert cartesian J2000 coordinates to geodetic coordinates.
    
    :param j2000Vecs: shape (n,3)
    :param datetime time_:
    :rtype: tuple (latitudes, longitudes) in degrees
    """
    t0 = time.time()
    j2000Vecs = np.asarray(j2000Vecs)
    geoX, geoY, geoZ = j2000_to_geo(time_, j2000Vecs).T
    print('convert J2000-GEO:', time.time()-t0, 's')
        
    t0 = time.time()
    lat, lon = ecef2Geodetic(geoX, geoY, geoZ, wgs84A, wgs84B)
    np.rad2deg(lat, lat)
    np.rad2deg(lon, lon)
    print('convert GEO-LatLon:', time.time()-t0, 's')
    
    return lat, lon

def latLonToJ2000(lat, lon, h, time_):
    """
    Convert geodetic coordinates to cartesian J2000 coordinates.
    
    :param lat: scalar or 1d-array, degrees
    :param lon: scalar or 1d-array, degrees
    :param h: scalar or 1d-array
    :param datetime time_:
    :rtype: tuple (latitudes, longitudes) in degrees
    """
    t0 = time.time()
    lat = np.deg2rad(lat)
    lon = np.deg2rad(lon)
    geoX, geoY, geoZ = geodetic2Ecef(lat, lon, h)
    geoVecs = np.asarray([geoX, geoY, geoZ]).T
    isScalar = geoVecs.ndim == 1
    if isScalar:
        geoVecs = np.array([geoVecs])
    print('convert LatLon-GEO:', time.time()-t0, 's')
        
    t0 = time.time()
    j2000 = geo_to_j2000(time_, geoVecs).T
    if isScalar:
        j2000 = j2000[:,0]
    print('convert GEO-J2000:', time.time()-t0, 's')
    
    return j2000

def smLonToMLT(smlons, out=None):
    """
    Convert solar magnetic longitudes to magnetic local time.
    
    :param smlons: in degrees [-180,180]
    :return: magnetic local time [0,24]
    """
    if out is not None:
        np.multiply(smlons, 24/360, out)
        mlt = out
    else:
        mlt = smlons*(24/360)
    mlt += 12
    return mlt

def mltToSmLon(mlt, out=None):
    """
    Convert magnetic local time to solar magnetic longitudes.
    
    :param mlt: in hours [0,24]
    :return: solar magnetic longitudes in degrees [-180,180]
    """
    if out is not None:
        np.subtract(mlt, 12, out)
        smlon = out
    else:
        smlon = mlt-12
    smlon /= (24/360)
    return smlon

def j2000ToMLatMLT(j2000Vecs, time_):
    """
    Convert cartesian J2000 coordinates to MLat/MLT coordinates using the
    IGRF model.
    
    MLat = geomagnetic latitude
    
    MLT = magnetic local time
    
    :param j2000Vecs: shape (n,3)
    :param datetime time_:
    :rtype: tuple (mlat, mlt) in (degrees,hours)
    """
    t0 = time.time()
    j2000Vecs = np.asarray(j2000Vecs)
    smX, smY, smZ = j2000_to_sm(time_, j2000Vecs).T
    print('convert J2000-SM:', time.time()-t0, 's')
    
    t0 = time.time()
    _, smlat, smlon = cartesian_to_spherical(smX, smY, smZ)
    np.rad2deg(smlat, smlat)
    np.rad2deg(smlon, smlon)
    mlat = smlat
    smLonToMLT(smlon, smlon)
    mlt = smlon
    print('convert SM-MLat/MLT:', time.time()-t0, 's')
    
    return mlat, mlt

def geoToMLatMLT(geoVecs, time_):
    """
    Convert ECEF coordinates to MLat/MLT coordinates using the
    IGRF model.
    
    MLat = geomagnetic latitude
    
    MLT = magnetic local time
    
    :param geoVecs: shape (n,3)
    :param datetime time_:
    :rtype: tuple (mlat, mlt) in (degrees,hours)
    """    
    t0 = time.time()
    geoVecs = np.asarray(geoVecs)
    smX, smY, smZ = geo_to_sm(time_, geoVecs).T
    print('convert GEO-SM:', time.time()-t0, 's')
    
    t0 = time.time()
    _, smlat, smlon = cartesian_to_spherical(smX, smY, smZ)
    np.rad2deg(smlat, smlat)
    np.rad2deg(smlon, smlon)
    mlat = smlat
    smLonToMLT(smlon, smlon)
    mlt = smlon
    print('convert SM-MLat/MLT:', time.time()-t0, 's')
    
    return mlat, mlt
    
def smToLatLon(smlats, smlons, time_):
    """
    Convert solar magnetic to geodetic coordinates using the
    IGRF model.
    
    :param smlats: in degrees [-90,90]
    :param smlons: in degrees [-180,180]
    :param time_:
    :rtype: tuple (latitudes, longitudes) in degrees
    """
    t0 = time.time()
    smlats, smlons = np.deg2rad(smlats), np.deg2rad(smlons)
    smX, smY, smZ = spherical_to_cartesian(1, smlats.ravel(), smlons.ravel())
    smVecs = np.array([smX,smY,smZ]).T
    geoX, geoY, geoZ = sm_to_geo(time_, smVecs).T
    print('convert SM-GEO:', time.time()-t0, 's')
    
    t0 = time.time()
    lats, lons = ecef2Geodetic(geoX, geoY, geoZ)
    np.rad2deg(lats, lats)
    np.rad2deg(lons, lons)
    lats = lats.reshape(smlats.shape)
    lons = lons.reshape(smlons.shape)
    print('convert GEO-LatLon:', time.time()-t0, 's')
    return lats, lons

# The following functions are adapted from cxform_manual.c of NASA's cxform library.
# The below versions are much faster as they pre-calculate and multiply multiple rotation matrices
# before converting coordinates.

# the following axis directions for gohlke's rotation_matrix match it with hapgood_matrix defined in cxform
X = [-1,0,0]
Y = [0,1,0]
Z = [0,0,-1]


def mag_lon(et):
    """
    Longitude of Earth's magnetic pole in radians
    """
    fracYearIndex = (et+3155803200.0)/157788000.0
    fracYear = fmod(fracYearIndex, 1.0)

    g11 = calcG11(fracYearIndex, fracYear)
    h11 = calcH11(fracYearIndex, fracYear)

    lambda0 = atan2(h11, g11) + pi
    return lambda0

def mag_lat(et):
    """
    Latitude of Earth's magnetic pole in radians
    """
    fracYearIndex = (et+3155803200.0)/157788000.0
    fracYear = fmod(fracYearIndex, 1.0)

    g01 = calcG01(fracYearIndex, fracYear)
    g11 = calcG11(fracYearIndex, fracYear)
    h11 = calcH11(fracYearIndex, fracYear)
    lambda0 = mag_lon(et)
  
    phi0 = pi/2 - atan((g11*cos(lambda0) + h11*sin(lambda0))/g01)
    return phi0

def date2es(date):
    """
    Converts UTC to ephemeris seconds.
    """
    jd = Time(date, scale='utc').jd
    return (jd - 2451545) * 86400
    # cxform used the following which limits precision unnecessarily to 1sec:
#    return long(round((jd - 2451545) * 86400))

def T0(et):
    """
    Julian Centuries from a certain time to 1 Jan 2000 12:00
    """
    return (et / 86400.0)/36525.0

def H(et):
    """
    time, in hours, since preceding UT midnight
    """
    jd    = (et / 86400.0) - 0.5
    dfrac = jd - int(jd)
    hh    = dfrac * 24.0

    if (hh < 0.0):
        hh += 24.0

    return hh

def lambda0(et):
    """
    Sun's ecliptic longitude in degrees
    """
    M = 357.528 + 35999.050 * T0(et)
    lambd = 280.460 + 36000.772 * T0(et)
    lambda0 = lambd + (1.915 - 0.0048 * T0(et)) * sin(np.deg2rad(M)) + 0.020 * sin(np.deg2rad(2 * M))
    return lambda0

def epsilon(et):
    """
    The obliquity of the ecliptic in degrees
    """
    return 23.439 - 0.013 * T0(et)

def mat_P(et):
    """
    J2000 to GEI matrix
    """
    t0 = T0(et)
    
    mat = rotation_matrix(np.deg2rad(-1.0*(0.64062 * t0  +  0.00030 * t0*t0)), Z)

    mat_tmp = rotation_matrix(np.deg2rad(0.55675 * t0  -  0.00012 * t0*t0), Y)
    mat = np.dot(mat, mat_tmp)
        
    mat_tmp = rotation_matrix(np.deg2rad(-1.0*(0.64062 * t0  +  0.00008 * t0*t0)), Z)
    mat = np.dot(mat, mat_tmp)
    return mat[:3,:3]

def mat_T1(et):
    """
    GEI to GEO matrix
    """
    theta = 100.461 + 36000.770 * T0(et) + 360.0*(H(et)/24.0)

    mat = rotation_matrix(np.deg2rad(theta), Z)
    return mat[:3,:3]

def mat_T2(et):
    """
    GEI to GSE matrix
    """
    mat = rotation_matrix(np.deg2rad(lambda0(et)), Z)
    mat_tmp = rotation_matrix(np.deg2rad(epsilon(et)), X)
    mat = np.dot(mat, mat_tmp)
    return mat[:3,:3]

def vec_Qe(et):
    """
    don't ask. [sic]
    """
    lat = mag_lat(et)
    lon = mag_lon(et)
    
    cos_lat = cos(lat)
    sin_lat = sin(lat)
    cos_lon = cos(lon)
    sin_lon = sin(lon)

    Qg = [cos_lat * cos_lon, cos_lat * sin_lon, sin_lat]

    mat = mat_T2(et)
    mat_tmp = mat_T1(et).T

    mat = np.dot(mat, mat_tmp)
    Qe = np.dot(mat, Qg)
    return Qe

def mat_T3(et):
    """
    GSE to GSM matrix
    """
    Qe = vec_Qe(et)
    psi = atan2(np.deg2rad(Qe[1]), np.deg2rad(Qe[2]))
    mat = rotation_matrix(-psi, X)
    return mat[:3,:3]

def mat_T4(et):
    """
    GSM to SM matrix
    """
    Qe = vec_Qe(et)
    mu = atan2(np.deg2rad(Qe[0]), np.deg2rad(sqrt(Qe[1]*Qe[1] + Qe[2]*Qe[2])))
    mat = rotation_matrix(-mu, Y)
    return mat[:3,:3]

def mat_T5(et):
    """
    GEO to MAG matrix
    """
    mat = rotation_matrix(mag_lat(et) - np.deg2rad(90.0), Y)
    mat_tmp = rotation_matrix(mag_lon(et), Z)
    mat = np.dot(mat, mat_tmp)
    return mat[:3,:3]

# def hapgood_matrix(theta, axis):
#     if axis==[1,0,0]:
#         axis = 0
#     elif axis==[0,1,0]:
#         axis = 1
#     elif axis==[0,0,1]:
#         axis = 2
# 
#     sin_theta = np.sin(theta)
#     cos_theta = np.cos(theta)
# 
#     t1 = (axis+1) % 3
#     t2 = (axis+2) % 3
#     if (t1 > t2):
#         tmp = t1
#         t1  = t2
#         t2  = tmp
#     
#     mat = np.zeros((3,3))
# 
#     mat[axis,axis] = 1.0
#     mat[t1,t1]     = cos_theta
#     mat[t2,t2]     = cos_theta
# 
#     mat[t1,t2]     =  sin_theta
#     mat[t2,t1]     = -sin_theta
#     return mat

# Note that the cxform functions of this module (including rotation_matrix) were 
# as a test fully implemented using sympy (arbitrary precision floats) to check 
# whether the repeated matrix multiplications (see below) had a significant effect
# on accuracy due to float rounding errors. The result was that the maximum error
# of any element in the rotation matrices was in the order of e-14.

def mat_j2000_to_geo(et):
    # j2000_twixt_gei -> gei_twixt_geo
    mat = np.dot(mat_T1(et), mat_P(et))
    return mat

def mat_j2000_to_sm(et):
    # j2000_twixt_gei -> gei_twixt_gse -> gse_twixt_gsm -> gsm_twixt_sm
    mat = mat_T4(et).dot(mat_T3(et)).dot(mat_T2(et)).dot(mat_P(et))
    return mat

def mat_geo_to_sm(et):
    # gei_twixt_geo (reverse) -> gei_twixt_gse -> gse_twixt_gsm -> gsm_twixt_sm
    mat = mat_T4(et).dot(mat_T3(et)).dot(mat_T2(et)).dot(mat_T1(et).T)
    return mat

def j2000_to_geo(date, vecsJ2000):
    return x_to_y(mat_j2000_to_geo, date, vecsJ2000)

def geo_to_j2000(date, vecsGeo):
    return x_to_y(mat_j2000_to_geo, date, vecsGeo, reverse=True)

def j2000_to_sm(date, vecsJ2000):
    return x_to_y(mat_j2000_to_sm, date, vecsJ2000)

def geo_to_sm(date, vecsGEO):
    return x_to_y(mat_geo_to_sm, date, vecsGEO)

def sm_to_geo(date, vecsSM):
    return x_to_y(mat_geo_to_sm, date, vecsSM, reverse=True)

def gei_to_geo(date, vecsGEI):
    return x_to_y(mat_T1, date, vecsGEI)

def geo_to_gei(date, vecsGEO):
    return x_to_y(mat_T1, date, vecsGEO, reverse=True)

def gei_to_gse(date, vecsGEI):
    return x_to_y(mat_T2, date, vecsGEI)

def gse_to_gsm(date, vecsGSE):
    return x_to_y(mat_T3, date, vecsGSE)

def gsm_to_sm(date, vecsGSM):
    return x_to_y(mat_T4, date, vecsGSM)

def x_to_y(matFn, date, vecs, reverse=False):
    vecs = np.asarray(vecs)
    assert vecs.ndim == 2 
    assert vecs.shape[1] == 3
        
    et = date2es(date)
    mat = matFn(et)
    if reverse:
        mat = mat.T
    vecsOut = matrix_multiply(mat, vecs[...,np.newaxis]).reshape(vecs.shape)
    return vecsOut

def northGeomagneticPoleLocation(date):
    """
    Calculate the approximate position of the north geomagnetic pole for the
    given date using the IGRF model.
    
    :param datetime date:
    :rtype: named (latitude, longitude) tuple, in degrees
    """
    # TODO are these geocentric or geodetic coordinates??
    et = date2es(date)
    lat, lon = mag_lat(et), mag_lon(et)
    lat = np.rad2deg(lat)
    lon = Angle(lon * u.rad).wrap_at(180 * u.deg).degree
    return Location(lat, lon)

__all__ = map(lambda f: f.__name__,
              [spherical_to_cartesian, cartesian_to_spherical,
               geodetic2Ecef, geodetic2EcefZero, ecef2Geodetic,
               j2000ToLatLon, latLonToJ2000,
               smLonToMLT, mltToSmLon, j2000ToMLatMLT, geoToMLatMLT, smToLatLon,
               northGeomagneticPoleLocation])
    
if __name__ == '__main__':
    from datetime import datetime
    d = datetime(2010,1,1)
    print('Geomagnetic pole at ' + str(d) + ':')
    print(northGeomagneticPoleLocation(d))
    


# Copyright European Space Agency, 2013

from functools import wraps

def lazy_property(fn):
    """
    Caches the result of a property.
    """
    attr_name = '_lazy_' + fn.__name__
    @property
    @wraps(fn)
    def _lazyprop(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)
    return _lazyprop

def inherit_docs(cls):
    """
    Inherits docstrings from base classes
    for all overridden methods and properties in `cls` which lack a docstring.
    """
    # see https://stackoverflow.com/a/23964187
    for name in dir(cls):
        func = getattr(cls, name)
        if func.__doc__: 
            continue
        for parent in cls.mro()[1:]:
            if not hasattr(parent, name):
                continue
            doc = getattr(parent, name).__doc__
            if not doc: 
                continue
            try:
                # __doc__'s of properties are read-only.
                # The work-around below wraps the property into a new property.
                if isinstance(func, property):
                    # We don't want to introduce new properties, therefore check
                    # if cls owns it or search where it's coming from.
                    # With that approach (using dir(cls) instead of var(cls))
                    # we also handle the mix-in class case.
                    wrapped = property(func.fget, func.fset, func.fdel, doc)
                    clss = list(filter(lambda c: name in vars(c).keys() and not getattr(c, name).__doc__, cls.mro()))
                    setattr(clss[0], name, wrapped)
                else:
                    try:
                        func = func.__func__ # for instancemethod's
                    except:
                        pass
                    func.__doc__ = doc
            except: # some __doc__'s are not writable
                pass
            break
    return cls


# Copyright European Space Agency, 2013

from __future__ import absolute_import, print_function

import unittest
from nose.plugins.attrib import attr
import os.path
from datetime import datetime

from auromat.mapping.iss import ISSMappingProvider
from auromat.resample import resample
try:
    from auromat.draw import drawStereographicMLatMLT, drawScanLinesMLatMLTCo,\
        saveFig
except ImportError as e:
    print(repr(e))
    # import is optional so that the test module import doesn't fail if
    # the tests are not actually run and matplotlib is not installed
    
from auromat.util.coroutine import broadcast

url = 'http://127.0.0.1:5001/api/georef_seqs/12'
cacheFolder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'iss_cache')

@attr('slow')
class Test(unittest.TestCase):
    def _testSingle(self):
        provider = ISSMappingProvider(url, cacheFolder, altitude=110)
        
        mapping = provider.get(datetime(2012,1,25,9,26,57))
        mapping = resample(mapping, arcsecPerPx=100)
        saveFig('test.png', drawStereographicMLatMLT(mapping))
        
    def testSequence(self):
        provider = ISSMappingProvider(url, cacheFolder, altitude=110)
        seq = provider.getSequence()
        broadcast(seq, drawScanLinesMLatMLTCo('test_scanlines.png'))
        


# Copyright European Space Agency, 2013

"""
Converts a downloaded dataset into a given format (CDF and netCDF currently)
and applies requested postprocessing steps like resampling.
"""

from __future__ import absolute_import, print_function
from six.moves import map
import argparse
from datetime import datetime
import os
import sys
from functools import partial

from auromat.mapping.themis import ThemisMappingProvider
from auromat.mapping.iss import ISSMappingProvider
from auromat.resample import resample, resampleMLatMLT
from auromat.util.os import makedirs

__all__ = ['main']

def date(datestring):
    return datetime.strptime(datestring, '%Y-%m-%dT%H:%M:%S')

description = '''
This tool converts downloaded data to a new format and optionally
applies postprocessing like resampling.
'''

epilog = '''
Examples:

Convert everything in the current directory with default settings:
auromat-convert --format cdf

Convert data in iss/seq-1 folder with 100km mapping altitude:
auromat-convert --data iss/seq-1 --format cdf --altitude 100

Resample and convert data:
auromat-convert --format cdf --resample --resolution 80

Don't store pixel corner coordinates:
auromat-convert --format cdf --without-bounds

Note that if you called auromat-download with --start and/or --end
dates, then you have to specify these here as well.
'''

class Grid(object):
    geo = 'geo'
    mag = 'mag'
    
class Format(object):
    cdf = 'cdf'
    netcdf = 'netcdf'

def getParser():
    parser = argparse.ArgumentParser(prog='auromat-convert', 
                                     epilog=epilog, description=description,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('--data', help='Data directory, by default the current directory. '
                                       'For ESA ISS sequences, this must be the subfolder '
                                       'corresponding to a single sequence, typically "seq-n". '
                                       'Under Unix-type shells, use "--data seq-*" to process '
                                       'all sequences.', 
                        default=os.getcwd())
    
    period = parser.add_argument_group('period', 
                                       'These arguments optionally specify which data to convert. '
                                       'If you used --start and/or --end in auromat-download, '
                                       'then the same values have to be used here as well.')
    period.add_argument('--start', help='UTC start date, format 2000-01-01T12:00:00', type=date)
    period.add_argument('--end', help='UTC end date (inclusive)', type=date)
    
    mappingArgs = parser.add_argument_group('mapping')
    mappingArgs.add_argument('--altitude', help='Altitude in km onto which to map the images, '
                                                'default is 110km', default=110, type=int)
    
    esaIssArgs = parser.add_argument_group('ESA ISS data')
    esaIssArgs.add_argument('--bps', help='bits per sample, default is 16', choices=[8,16], default=16)
    esaIssArgs.add_argument('--correctgamma', 
                            help='Applies BT.709 gamma correction for creating visually '
                                 'pleasing images. A nonlinear curve is applied which '
                                 'brightens dark pixels more than bright ones to match '
                                 'human perception. If --correctgamma is not used then '
                                 'a more scientifically useful linear image is produced '
                                 'where the number of photons hitting a pixel is in linear '
                                 'relationship with the resulting pixel value.', 
                            action='store_true')
    esaIssArgs.add_argument('--autobright', 
                            help='Automatically brightens the image such that 1%% of all pixels '
                                 'are fully saturated. This may destroy information in the image '
                                 'and should not be used when doing a scientific analysis.',
                            action='store_true')
    
    resampleArgs = parser.add_argument_group('resampling')
    resampleArgs.add_argument('--resample', help='Whether to resample or not', action='store_true')
    resampleArgs.add_argument('--resolution', metavar='RES', help='in arcsec/px, default 100', default=100, type=float)
    resampleArgs.add_argument('--grid', 
                              help='The grid which will be regular after resampling. '
                                   'Default is MLat/MLT grid. Use geo for geographical grid.', 
                              default=Grid.mag, choices=[Grid.geo, Grid.mag])

    outputArgs = parser.add_argument_group('output')
    outputArgs.add_argument('--out', 
                            help='Output directory, by default the "converted" '
                                 'subdirectory of --data')
    outputArgs.add_argument('--overwrite', 
                            help='Overwrites existing files.',
                            action='store_true')
    outputArgs.add_argument('--skip',
                            help='Skips already converted files.',
                            action='store_true')
    outputArgs.add_argument('--format', help='Data format of converted files', 
                            choices=[Format.cdf, Format.netcdf], required=True)
    outputArgs.add_argument('--without-bounds', dest='withoutBounds',
                            help='Do not include coordinates of pixel corners. '
                                 'If set, then only the pixel center coordinates are written, '
                                 'otherwise both.',
                            action='store_true')
    outputArgs.add_argument('--without-mag', dest='withoutMag',
                            help='Do not include MLat/MLT coordinates.', 
                            action='store_true')
    outputArgs.add_argument('--without-geo', dest='withoutGeo',
                            help='Do not include geodetic coordinates. Only usable with CDF output.', 
                            action='store_true')
    
    parser.add_argument('--version', action='version', version='auromat TODO')
    return parser

def parseargs():
    parser = getParser()
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
        
    args = parser.parse_args()
    if not args.out:
        args.out = os.path.join(args.data, 'converted')
    if args.overwrite and args.skip:
        parser.error('only one of --overwrite and --skip is allowed')
    if args.withoutGeo and args.format == Format.netcdf:
        parser.error('--without-geo is only usable with --format cdf')
    
    return args

def main():
    args = parseargs()
        
    # first, we try to figure out from which provider the data came from
    dataPath = args.data
    dataFiles = os.listdir(dataPath)
    if any(f == 'api.json' for f in dataFiles):
        provider = ISSMappingProvider(dataPath, altitude=args.altitude,
                                      noRawPostprocessCaching=True,
                                      raw_bps=args.bps, raw_auto_bright=args.autobright,
                                      raw_gamma=None if args.correctgamma else (1,1),
                                      offline=True)
    
    elif any(f.startswith('thg_l1_') for f in dataFiles):
        if args.format == Format.cdf:
            print ('Note that THEMIS files are already in CDF format where each file '
                   'contains 1h of images. With this script a CDF file for each image '
                   'is created.')
        if not (args.start and args.end):
            print('For THEMIS data you have to specify --start and --end')
            sys.exit(1)
                        
        provider = ThemisMappingProvider(dataPath, dataPath, altitude=args.altitude,
                                         offline=True)
    
    else:
        raise NotImplementedError('Not recognized as THEMIS or ESA ISS data')
    
    mappings = provider.getSequence(args.start, args.end)
    
    if args.resample:
        if args.grid == Grid.geo:
            resample_ = resample
        elif args.grid == Grid.mag:
            resample_ = resampleMLatMLT
        resample_ = partial(resample_, arcsecPerPx=args.resolution)
            
        mappings = map(resample_, mappings)
    
    if args.format == Format.cdf:
        import auromat.export.cdf
        export = auromat.export.cdf.write
        ext = '.cdf'
        
    elif args.format == Format.netcdf:
        import auromat.export.netcdf
        export = auromat.export.netcdf.write
        ext = '.nc'
    
    export = partial(export, includeBounds=not args.withoutBounds, includeMagCoords=not args.withoutMag,
                     includeGeoCoords=not args.withoutGeo)
        
    makedirs(args.out)
    for mapping in mappings:
        path = os.path.join(args.out, mapping.identifier + ext)
        if os.path.exists(path):
            if args.skip:
                print('skipping', path)
                continue
            elif args.overwrite:
                os.remove(path)
            else:
                print('The file', path, 'already exists.\n'
                      'Please use --skip or --overwrite, or a different output folder.',
                      file=sys.stderr)
                sys.exit(1)
            
        print('storing', path)
        export(path, mapping)
    
    print('Done.')

main.__doc__ = """
::
  
  {}
  

""".format(getParser().format_help().replace('\n', '\n  '))

if __name__ == '__main__':
    main()
    


"""
The auromat package is split up in several packages and modules each covering
different aspects of its functionality.

The :mod:`auromat.cli` package contains command-line tools that will be installed as `auromat-name`
for a module name `auromat.cli.name`. It is not intended to be used from within
Python code.

The :mod:`auromat.coordinates` package allows to determine coordinates of space objects, 
to convert existing coordinates into other reference frames,
to calculate intersection points between a ray and an ellipsoid, and to perform geodesic
calculations. It does not depend on the `mapping` objects used in other parts of
the :mod:`auromat` package and can therefore be easily re-used for other purposes.

The :mod:`auromat.solving` package contains modules to blindly georeference
astronaut photography pointing to earth using the starfield in the images.
The use of this package is not required for working with already georeferenced
images such as the ones from the ESA-ISS archive.

The :mod:`auromat.mapping` package contains various modules for reading
and working with georeferenced data, e.g. as produced by the :mod:`auromat.solving`
package or other available in forms like the THEMIS or ESA-ISS web archives.
It also contains the
:class:`~auromat.mapping.mapping.BaseMapping` main class which other packages depend
on for accessing georeferenced image data through a defined interface.

The :mod:`auromat.export` package can export any :class:`~auromat.mapping.mapping.BaseMapping`
into a self-contained CDF or netCDF file, suitable for further processing in different
software.

The :mod:`auromat.util` package contains independent generic helper functions not strictly related
to the main functions of this library. No module has a dependency to another part of
this library.

The :mod:`auromat` package also contains various submodules, e.g. for 
resampling (:mod:`auromat.resample`) and visualizing (:mod:`auromat.draw`)
mappings, as well as some modules which did not fit well into any other bigger package.
"""

from __future__ import absolute_import

from ._version import __version__, __version_info__

import warnings
try:
    import matplotlib as mpl
    # Agg is the only backend supporting the rasterized option so we can't just use cairo.
    mpl.use('Agg') # headless backend
except ImportError:
    pass
else:
    from distutils.version import LooseVersion
    if LooseVersion(mpl.__version__) < '1.4.0':
        warnings.warn('Your matplotlib version has a memory leak which occurs when saving maps ' +
                      'in SVG format. Use >= 1.4.0 if you need to draw many maps in a row. ' +
                      'More details are at https://github.com/matplotlib/matplotlib/issues/3197')


# Copyright European Space Agency, 2013

from __future__ import print_function

from six import reraise, PY3
from functools import wraps
import sys

# see http://slideshare.net/emptysquare/nyc-python-meetup-coroutines-2013-0416

def coroutine(func):
    @wraps(func)
    def start(*args,**kwargs):
        cr = func(*args,**kwargs)
        next(cr)
        return cr
    return start

def broadcast(iterable, *targets):
    """
    Sends items to multiple coroutines.
    
    If one coroutine raises an exception, then all other coroutines
    are killed as well and the exception is re-raised to the caller.
    
    :param iterable: the items to send to the coroutines
    :param targets: coroutines to send iterables to
    """
    _sendTo(iterable, _broadcast(targets))

@coroutine
def _broadcast(targets):
    exc = None
    
    if len(targets) == 1:
        # optimized code path which doesn't hold a reference
        # to the yielded item -> conserves memory if item
        # is replaced/transformed in target
        target = targets[0]
        while True:
            try:
                try:
                    target.send((yield))
                except GeneratorExit:
                    break
                except StopIteration:
                    pass
                except:
                    exc = sys.exc_info()
                    raise GeneratorExit
            except GeneratorExit:
                break
    else:
        while True:
            try:
                item = (yield)
                for target in targets:
                    try:
                        target.send(item)
                    except (GeneratorExit, StopIteration):
                        pass
                    except:
                        exc = sys.exc_info()
                        raise GeneratorExit
            except GeneratorExit:
                break
            except:
                exc = sys.exc_info()
                break
        
        # in case item is big, remove reference to it
        # to conserve memory for the closing phase
        try:
            del item
        except NameError:
            pass
        
    if exc is None:
        for target in targets:
            try:
                target.close()
            except (GeneratorExit, StopIteration):
                pass
            except:
                exc = sys.exc_info()
                break
    
    if exc is not None:
        etype, e, tb = exc
        
        # kill other coroutines
        for target in targets:
            try:
                throw(target, etype, e, tb)
            except:
                pass
        
        reraise(etype, e, tb)
         
def _sendTo(iterable, target):
    try:
        try:
            while True:
                # important: we use next() so that we don't have
                # to store a reference in a local variable in a for-loop
                # (see _broadcast())
                target.send(next(iterable))
        except StopIteration:
            pass
        target.close()
    except:
        # if generator raised, then we kill the target with the same exception
        etype, ei, tb = sys.exc_info()
        throw(target, etype, ei, tb) # will raise back to us

def throw(target, etype, e, tb):
    if PY3:
        target.throw(etype(e).with_traceback(tb))
    else:
        target.throw(etype, e, tb)

# Example use
if __name__ == '__main__':
    from time import sleep
    
    @coroutine
    def grep(pattern, e):
        print("Looking for %s" % pattern)
        try:
            line = yield
            line = yield
        except GeneratorExit:
            raise RuntimeError('not enough items!')
                    
        while True:
            sleep(1)
            try:
                line = yield
            except GeneratorExit:
                print('close was called, no more items left')
                break
            
            if pattern in line:
                print(line)
                
            if e:
                raise ValueError('I am a bad consumer')
         
        print('end')

    # main limitation: no single return value (only for each item)

    def gen():
        for i in range(10):
            yield str(i)
            if i == 5:
                raise ValueError('I am a bad generator')

    consumers = [grep('foo', True), grep('bar', False)]
    broadcast(["fooooo","barrrr"]*3, consumers)
    #broadcast(["fooooo"], consumers)
    #broadcast(gen(), consumers)

    sleep(5)
    print('program exit')


# Copyright European Space Agency, 2013

from __future__ import absolute_import

import os
import errno

def touch(path):
    """
    Create new file if it doesn't exist or just update mtime to NOW.
    """
    with open(path, 'a'):
        os.utime(path, None)
        
def makedirs(*paths):
    """
    Recursively creates folders if not already existing.
    """
    for path in paths:
        try:
            os.makedirs(path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise


# Copyright European Space Agency, 2013

from __future__ import division

import unittest
import os

import auromat.util.lensdistortion

class Test(unittest.TestCase):
    def test(self):
        imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
        auromat.util.lensdistortion.correctLensDistortion(imagePath, 'test_undist.jpg', preserveExif=False)
        
def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)



# Copyright European Space Agency, 2013

from setuptools import setup, find_packages
import re

# version handling from https://stackoverflow.com/a/7071358
VERSIONFILE="auromat/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(
    name = 'auromat',
    description = 'AUROra MApping Toolkit',
    long_description = open('README.rst').read(),
    version = verstr,
    author = 'Maik Riechert',
    author_email = 'mriecher@cosmos.esa.int, awalsh@sciops.esa.int',
    license = 'ESCL - Type 1',
    url = 'https://github.com/esa/auromat',
    classifiers=[
      'Development Status :: 5 - Production/Stable',
      'Intended Audience :: Science/Research',
      'Natural Language :: English',
      'Programming Language :: Python :: 2',
      'Programming Language :: Python :: 3',
      'Operating System :: OS Independent',
      'Topic :: Scientific/Engineering',
      'Topic :: Software Development :: Libraries',
    ],
    packages = find_packages(),
    install_requires=['six',
                      'futures',
                      'papy',
                      'numpy>=1.6',
                      'scipy>=0.9',
                      'matplotlib', # required by auromat.utils
                      'brewer2mpl',
                      'numexpr',
                      'scikit-image',
                      'pillow', # used by scikit-image
                      'psutil',
                      'ExifRead',
                      'geographiclib',
                      'astropy>=0.4.1',
                      'astroquery',
                      'pyephem>=3.7.5.2',
                      'lensfunpy>=1.1.0',
                      'rawpy>=0.1',
                      ],
    extras_require = {
        'netcdf': ['netCDF4'], # for netCDF export
        'cdf': ['spacepy'], # for CDF export and reading THEMIS CDF files
    },
    package_data = {
        '': ['*.dbf', '*.shp', '*.shx'],
    },
    entry_points = {
        'console_scripts': [
            'auromat-download = auromat.cli.download:main',
            'auromat-convert = auromat.cli.convert:main',
        ]
    }
)


# Copyright European Space Agency, 2013

from __future__ import print_function

import unittest
from nose.plugins.attrib import attr
import os
import numpy as np
from auromat.mapping.spacecraft import getMapping
from auromat.resample import resample
from auromat.mapping import miracle

@attr('slow')
class Test(unittest.TestCase):

    def testSpacecraftMappingNorth2(self):
        # late import so that the test module import doesn't fail if
        # the tests are not actually run and matplotlib is not installed
        from auromat.draw_helpers import generatePolygonsFromMapping
        
        m = _getMappingNorth()
        foo = generatePolygonsFromMapping(m)
        print(foo)

    def testSpacecraftMappingNorth(self):
        m = _getMappingNorth()
        m.checkGuarantees()
        m2 = m.maskedByElevation(10)
        m2.checkGuarantees()
        
        assert np.any(~(m.latsCenter.mask == m2.latsCenter.mask))
        assert np.any(~(m.lats.mask == m2.lats.mask))
        
        m3 = resample(m, arcsecPerPx=100, method='mean')
        m3.checkGuarantees()
        
    def testSpacecraftMappingSouth(self):
        m = _getMappingSouth()
        m.checkGuarantees()
        m = m.maskedByElevation(10)
        m.checkGuarantees()
        m = resample(m, arcsecPerPx=100, method='mean')
        m.checkGuarantees()
        
    def testMiracleMapping(self):
        m = _getMiracleMapping()
        m.checkGuarantees()
        m = m.maskedByElevation(10)
        m.checkGuarantees()
        m = resample(m, arcsecPerPx=100, method='mean')
        m.checkGuarantees()

def getResourcePath(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/" + name)

def _getMappingNorth():
    imagePath = getResourcePath('ISS030-E-102170_dc.jpg')
    wcsPath = getResourcePath('ISS030-E-102170_dc.wcs')
    mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    return mapping

def _getMappingSouth():
    imagePath = getResourcePath('ISS029-E-8492.jpg')
    wcsPath = getResourcePath('ISS029-E-8492.wcs')
    mapping = getMapping(imagePath, wcsPath, fastCenterCalculation=True)
    return mapping

def _getMiracleMapping():
    imagePath = getResourcePath('SOD120304_171900_557_1000.jpg')
    mapping = miracle.getMapping(imagePath)
    return mapping
